from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, electric_power_conservation_test, feasibility_test


class TestMILPElectricSourceSink(TestCase):
    def test_source_sink(self):
        """
        Tests for an electricity network that consist out of a source, a sink and storage,
        connected with electricity cables.

        Checks:
        - Check that the efficiency for charging and discharging is used proper to determine the
        consumed and stored power
        - Check that the battery is used to match demand
        - Check that charging and discharging properly connected to the network (direction)
        - Check that the is_charging variable is set correctly


        """

        import models.unit_cases_electricity.battery.src.example as example
        from models.unit_cases_electricity.battery.src.example import ElectricityProblem

        base_folder = Path(example.__file__).resolve().parent.parent
        tol = 1e-10

        solution = run_esdl_mesido_optimization(
            ElectricityProblem,
            base_folder=base_folder,
            esdl_file_name="source_sink_storage.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        results = solution.extract_results()
        parameters = solution.parameters(0)

        feasibility_test(solution)

        demand_matching_test(solution, results)
        electric_power_conservation_test(solution, results)

        storage_name = solution.energy_system_components.get("electricity_storage")[0]
        charge_eff = parameters[f"{storage_name}.charge_efficiency"]
        discharge_eff = parameters[f"{storage_name}.discharge_efficiency"]
        is_charging = results[f"{storage_name}__is_charging"]
        eff_power_change_bat = results[f"{storage_name}.Effective_power_charging"]
        eff_power_change_discharge_bat = results[f"{storage_name}__effective_power_discharging"]
        power_bat_network = results[f"{storage_name}.ElectricityIn.Power"]
        stored_el = results[f"{storage_name}.Stored_electricity"]

        power_cable_bat = results["ElectricityCable_91c1.ElectricityOut.Power"]
        np.testing.assert_allclose(power_cable_bat, power_bat_network, atol=tol)

        # if battery is charging (1), ElectricityIn.Power and effective_power charging should be
        # positive, else negative
        bigger_then = all(is_charging * eff_power_change_bat >= 0)
        smaller_then = all((1 - is_charging) * eff_power_change_bat <= 0)
        self.assertTrue(bigger_then)
        self.assertTrue(smaller_then)

        bigger_then = all(is_charging * power_bat_network >= 0)
        smaller_then = all((1 - is_charging) * power_bat_network <= 0)
        self.assertTrue(bigger_then)
        self.assertTrue(smaller_then)

        # battery should be charging at atleast one timestep to overcome the difference between max
        # production and demand
        self.assertTrue(sum(is_charging) >= 1)

        # stored electricity starts at 0
        np.testing.assert_allclose(
            stored_el[0],
            0.0,
            err_msg="The battery should be empty at the start to check if it operates and "
            "predicted",
        )

        # stored electricity change should be equal to the effective power change
        stored_change = stored_el[1:] - stored_el[:-1]
        np.testing.assert_allclose(eff_power_change_bat[1:] * 3600, stored_change, atol=tol)

        # effective power change while charging should be equal to efficiency * powerIn
        np.testing.assert_allclose(
            eff_power_change_bat * is_charging,
            power_bat_network * is_charging * charge_eff,
            atol=tol,
        )
        # effective power change while discharging should be equal to powerOut/efficiency
        np.testing.assert_allclose(
            eff_power_change_bat * (1 - is_charging),
            power_bat_network * (1 - is_charging) / discharge_eff,
            atol=tol,
        )

        # effective power discharge variable should always be the bigger or equal to the negative
        # of the charge variable and bigger then zero. The zero should only occur if battery is
        # charging. When a goal would be set to minimise discharge it should match the charge power,
        # however now this goal is not turned on.
        # TODO: when the new goal is included create test, this will end up in the mc_simulator
        np.testing.assert_array_less(-eff_power_change_bat, eff_power_change_discharge_bat + tol)
        self.assertTrue(all(eff_power_change_discharge_bat >= 0.0))



from mesido.pycml import Variable

from numpy import nan

from .gas_base import GasPort
from .._internal import BaseAsset
from .._internal.gas_component import GasComponent


class GasDemand(GasComponent, BaseAsset):
    """
    A gas demand consumes flow from the network.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "gas_demand"
        self.min_head = 30.0

        self.Q_nominal = nan

        self.density = 2.5e3  # H2 density [g/m3] at 30bar

        self.id_mapping_carrier = -1

        self.add_variable(GasPort, "GasIn")
        self.add_variable(
            Variable, "Gas_demand_mass_flow", min=0.0, nominal=self.Q_nominal * self.density
        )

        self.add_equation(
            ((self.GasIn.mass_flow - self.Gas_demand_mass_flow) / (self.Q_nominal * self.density))
        )



import datetime
import json
import logging
import math
import numbers
import os
import sys
import traceback
import uuid
from pathlib import Path

import esdl
from esdl.profiles.influxdbprofilemanager import ConnectionSettings
from esdl.profiles.influxdbprofilemanager import InfluxDBProfileManager
from esdl.profiles.profilemanager import ProfileManager

import mesido.esdl.esdl_parser
from mesido.constants import GRAVITATIONAL_CONSTANT
from mesido.esdl.edr_pipe_class import EDRPipeClass
from mesido.workflows.utils.helpers import _sort_numbered

import numpy as np

import pandas as pd


from rtctools.optimization.timeseries import Timeseries


logger = logging.getLogger("mesido")


class ScenarioOutput:
    __optimized_energy_system_handler = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.model_folder = kwargs.get("model_folder")
        self.output_folder = kwargs.get("output_folder")
        self.esdl_file_name = kwargs.get("esdl_file_name", "ESDL_file.esdl")
        # Settings for influxdb when writing out result profile data to it
        # Default settings
        self.write_result_db_profiles = False
        self.influxdb_username = None
        self.influxdb_password = None

        base_error_string = "Missing influxdb setting for writing result profile data:"
        try:
            self.write_result_db_profiles = kwargs["write_result_db_profiles"]

            if self.write_result_db_profiles not in [True, False]:
                logger.error(
                    "Current setting of write_result_db_profiles is: "
                    f"{self.write_result_db_profiles} and it should be set to True or False"
                )
                sys.exit(1)

            if self.write_result_db_profiles:
                try:
                    self.influxdb_host = kwargs["influxdb_host"]
                    if len(self.influxdb_host) == 0:
                        logger.error(
                            "Current setting of influxdb_host is an empty string and it should"
                            " be the name of the host"
                        )
                        sys.exit(1)
                except KeyError:
                    logger.error(f"{base_error_string} host")
                    sys.exit(1)
                try:
                    self.influxdb_port = kwargs["influxdb_port"]
                    if not isinstance(self.influxdb_port, int):
                        logger.error(
                            "Current setting of influxdb_port is: "
                            f"{self.influxdb_port} and it should be set to int value (port number)"
                        )
                        sys.exit(1)
                except KeyError:
                    logger.error(f"{base_error_string} port")
                    sys.exit(1)
                try:
                    self.influxdb_username = kwargs["influxdb_username"]
                except KeyError:
                    logger.error(f"{base_error_string} username")
                    sys.exit(1)
                try:
                    self.influxdb_password = kwargs["influxdb_password"]
                except KeyError:
                    logger.error(f"{base_error_string} password")
                    sys.exit(1)
                try:
                    self.influxdb_ssl = kwargs["influxdb_ssl"]
                    if self.influxdb_ssl not in [True, False]:
                        logger.error(
                            "Current setting of influxdb_ssl is: "
                            f"{self.influxdb_ssl} and it should be set to True or False"
                        )
                        sys.exit(1)
                except KeyError:
                    logger.error(f"{base_error_string} ssl")
                    sys.exit(1)
                try:
                    self.influxdb_verify_ssl = kwargs["influxdb_verify_ssl"]
                    if self.influxdb_verify_ssl not in [True, False]:
                        logger.error(
                            "Current setting of influxdb_verify_ssl is: "
                            f"{self.influxdb_verify_ssl} and it should be set to True or False"
                        )
                        sys.exit(1)
                except KeyError:
                    logger.error("f{base_string} verify_ssl")
                    sys.exit(1)
        except KeyError:
            # Not writing out to a influxdb, so no settings are requried
            pass

    def get_optimized_esh(self):
        return self.__optimized_energy_system_handler

    def _write_html_output(self, template_name="mpc_buffer_sizing_output"):
        from jinja2 import Environment, FileSystemLoader

        assert self.ensemble_size == 1

        results = self.extract_results()
        parameters = self.parameters(0)

        # Format the priority results
        priority_results = [
            dict(
                number=number,
                success=success,
                pretty_time=f"{int(seconds // 60):02d}:{seconds % 60:06.3f}",
                objective_value=objective_value,
                return_status=stats["return_status"],
                secondary_return_status=stats.get("secondary_return_status", ""),
            )
            for (
                number,
                seconds,
                success,
                objective_value,
                stats,
            ) in self._priorities_output
        ]

        # Format the buffer results
        results_buffers_placed = {}
        results_buffers_size = {}
        results_sources_placed = {}
        results_sources_size = {}
        results_max_charging_rate = {}
        results_max_discharging_rate = {}

        for buffer in _sort_numbered(self.energy_system_components.get("heat_buffer", [])):
            if buffer in self._minimize_size_buffers:
                max_size_var = self._max_buffer_heat_map[buffer]
                results_buffers_size[buffer] = float(results[max_size_var][0]) / (
                    parameters[f"{buffer}.cp"]
                    * parameters[f"{buffer}.rho"]
                    * (parameters[f"{buffer}.T_supply"] - parameters[f"{buffer}.T_return"])
                )
            else:
                results_buffers_size[buffer] = "-"

            if buffer in self._optional_buffers:
                buffer_placement_var = self._buffer_placement_map[buffer]
                results_buffers_placed[buffer] = np.round(results[buffer_placement_var][0]) == 1.0
            else:
                results_buffers_placed[buffer] = "-"

            (_, hot_orient), _ = self.energy_system_topology.buffers[buffer]
            q = hot_orient * results[f"{buffer}.HeatIn.Q"]
            inds_charging = q > 0
            inds_discharging = q < 0

            results_max_charging_rate[buffer] = max(q[inds_charging]) if any(inds_charging) else 0.0
            results_max_discharging_rate[buffer] = (
                max(-1 * q[inds_discharging]) if any(inds_discharging) else 0.0
            )

        buffer_results = [
            dict(
                name=buffer,
                tune_size=buffer in self._minimize_size_buffers,
                tune_placement=buffer in self._optional_buffers,
                maximum_size=self._override_buffer_size[buffer],
                result_placed=results_buffers_placed[buffer],
                result_size=results_buffers_size[buffer],
                max_charging_rate=results_max_charging_rate[buffer],
                max_discharging_rate=results_max_discharging_rate[buffer],
            )
            for buffer in self.energy_system_components.get("heat_buffer", [])
        ]

        for source in _sort_numbered(self.energy_system_components["heat_source"]):
            if source in self._minimize_size_sources:
                max_size_var = self._max_source_heat_map[source]
                results_sources_size[source] = float(results[max_size_var][0]) / 10.0**3
            else:
                results_sources_size[source] = "-"

            if source in self._optional_sources:
                source_placement_var = self._source_placement_map[source]
                results_sources_placed[source] = np.round(results[source_placement_var][0]) == 1.0
            else:
                results_sources_placed[source] = "-"

        source_results = [
            dict(
                name=source,
                tune_size=source in self._minimize_size_sources,
                tune_placement=source in self._optional_sources,
                maximum_size=self._override_max_power[source],
                result_placed=results_sources_placed[source],
                result_size=results_sources_size[source],
            )
            for source in self.energy_system_components["heat_source"]
        ]

        # Format the pipe results
        # Note that we do not distinguish between routing and sizing
        # internally, but for the sake of the output we do.
        pipe_results = []

        for p in _sort_numbered(self.hot_pipes):
            pipe_classes = self.pipe_classes(p)
            tune_routing = len([pc for pc in pipe_classes if pc.inner_diameter == 0.0]) == 1
            inner_diameter = parameters[f"{p}.diameter"]
            asset = next(a for a in self.esdl_assets.values() if a.name == p)
            esdl_diameter = asset.attributes["innerDiameter"]

            if len(pipe_classes) <= 1:
                tune_size = False
                min_dn_size = "-"
                max_dn_size = "-"
                result_placed = "-"
                result_dn_size = "-"
            elif len(pipe_classes) == 2 and tune_routing:
                tune_size = False
                min_dn_size = "-"
                max_dn_size = "-"
                result_placed = "Yes" if inner_diameter > 0 else "No"
                result_dn_size = "-"
            else:
                sorted_pipe_classes = sorted(
                    [pc for pc in pipe_classes if pc.inner_diameter > 0],
                    key=lambda pc: pc.inner_diameter,
                )

                tune_size = True
                min_dn_size = sorted_pipe_classes[0].name
                max_dn_size = sorted_pipe_classes[-1].name
                result_placed = "Yes" if inner_diameter > 0 else "No"
                result_pipe_class = self.get_optimized_pipe_class(p)
                result_dn_size = (
                    result_pipe_class.name if result_pipe_class is not None else inner_diameter
                )

            pipe_results.append(
                dict(
                    name=p,
                    tune_size=tune_size,
                    tune_routing=tune_routing,
                    esdl_diameter=esdl_diameter,
                    min_dn_size=min_dn_size,
                    max_dn_size=max_dn_size,
                    result_placed=result_placed,
                    result_dn_size=result_dn_size,
                )
            )

        input_csv_tables = {
            os.path.basename(x): pd.read_csv(x).to_dict("records")
            for x in self._csv_input_parameter_files
        }

        # Actually write out the html file based on the template
        templates_dir = Path(__file__).parent / "templates"

        env = Environment(loader=FileSystemLoader(templates_dir))
        template = env.get_template(template_name + ".html")

        os.makedirs(self._html_output_dir, exist_ok=True)

        filename = self._html_output_dir / (template_name + ".html")

        with open(filename, "w", encoding="utf-8") as fh:
            fh.write(
                template.render(
                    buffer_results=buffer_results,
                    source_results=source_results,
                    pipe_results=pipe_results,
                    priority_results=priority_results,
                    input_csv_tables=input_csv_tables,
                )
            )

    def _add_kpis_to_energy_system(self, energy_system, optimizer_sim: bool = False):

        results = self.extract_results()
        parameters = self.parameters(0)

        # ------------------------------------------------------------------------------------------
        # KPIs
        # General cost breakdowns
        # ------------------------------------------------------------------------------------------
        kpis_top_level = esdl.KPIs(id=str(uuid.uuid4()))
        heat_source_energy_wh = {}
        asset_opex_breakdown = {}  # yearly cost
        tot_variable_opex_cost_euro = 0.0  # yearly cost
        tot_fixed_opex_cost_euro = 0.0  # yearly cost

        # cost over the total time horizon (number of year) being optimized:
        #   - parameters["number_of_years"]
        #   - these kpis are not created for the newtwork simualtor->optimizer_sim
        asset_timehorizon_opex_breakdown = {}
        tot_timehorizon_variable_opex_cost_euro = 0.0
        tot_timehorizon_fixed_opex_cost_euro = 0.0
        asset_timehorizon_capex_breakdown = {}
        tot_timehorizon_install_cost_euro = 0.0
        tot_timehorizon_invest_cost_euro = 0.0

        # Specify the correct time horizon:
        # Optimization=number of year: since it is taken into account in TCO minimization
        # Simulator=1: since 30 years of optimization is not applicable for the network simulator
        if not optimizer_sim:  # optimization mode
            optim_time_horizon = parameters["number_of_years"]
        elif optimizer_sim:  # network simulator mode
            optim_time_horizon = 1.0
        else:
            logger.error("Variable optimizer_sim has not been set")

        for _key, asset in self.esdl_assets.items():
            asset_placement_var = self._asset_aggregation_count_var_map[asset.name]
            placed = np.round(results[asset_placement_var][0]) >= 1.0

            if np.isnan(parameters[f"{asset.name}.technical_life"]) or np.isclose(
                parameters[f"{asset.name}.technical_life"], 0.0
            ):
                capex_factor = 1.0
            else:
                capex_factor = math.ceil(
                    optim_time_horizon / parameters[f"{asset.name}.technical_life"]
                )

            if placed:
                try:
                    asset_timehorizon_capex_breakdown[asset.asset_type] += (
                        results[f"{asset.name}__installation_cost"][0]
                        + results[f"{asset.name}__investment_cost"][0]
                    ) * capex_factor
                    tot_timehorizon_install_cost_euro += (
                        results[f"{asset.name}__installation_cost"][0]
                    ) * capex_factor
                    tot_timehorizon_invest_cost_euro += (
                        results[f"{asset.name}__investment_cost"][0]
                    ) * capex_factor

                    if (
                        results[f"{asset.name}__variable_operational_cost"][0] > 0.0
                        or results[f"{asset.name}__fixed_operational_cost"][0] > 0.0
                    ):
                        asset_opex_breakdown[asset.asset_type] += (
                            results[f"{asset.name}__variable_operational_cost"][0]
                            + results[f"{asset.name}__fixed_operational_cost"][0]
                        )
                        asset_timehorizon_opex_breakdown[asset.asset_type] += (
                            results[f"{asset.name}__variable_operational_cost"][0]
                            + results[f"{asset.name}__fixed_operational_cost"][0]
                        ) * optim_time_horizon

                        tot_variable_opex_cost_euro += results[
                            f"{asset.name}__variable_operational_cost"
                        ][0]
                        tot_fixed_opex_cost_euro += results[
                            f"{asset.name}__fixed_operational_cost"
                        ][0]
                        tot_timehorizon_variable_opex_cost_euro += (
                            results[f"{asset.name}__variable_operational_cost"][0]
                            * optim_time_horizon
                        )
                        tot_timehorizon_fixed_opex_cost_euro += (
                            results[f"{asset.name}__fixed_operational_cost"][0] * optim_time_horizon
                        )

                except KeyError:
                    try:
                        asset_timehorizon_capex_breakdown[asset.asset_type] = (
                            results[f"{asset.name}__installation_cost"][0]
                            + results[f"{asset.name}__investment_cost"][0]
                        ) * capex_factor
                        tot_timehorizon_install_cost_euro += (
                            results[f"{asset.name}__installation_cost"][0]
                        ) * capex_factor
                        tot_timehorizon_invest_cost_euro += (
                            results[f"{asset.name}__investment_cost"][0]
                        ) * capex_factor

                        if (
                            results[f"{asset.name}__variable_operational_cost"][0] > 0.0
                            or results[f"{asset.name}__fixed_operational_cost"][0] > 0.0
                        ):
                            asset_opex_breakdown[asset.asset_type] = (
                                results[f"{asset.name}__variable_operational_cost"][0]
                                + results[f"{asset.name}__fixed_operational_cost"][0]
                            )
                            asset_timehorizon_opex_breakdown[asset.asset_type] = (
                                asset_opex_breakdown[asset.asset_type] * optim_time_horizon
                            )

                            tot_variable_opex_cost_euro += results[
                                f"{asset.name}__variable_operational_cost"
                            ][0]
                            tot_fixed_opex_cost_euro += results[
                                f"{asset.name}__fixed_operational_cost"
                            ][0]
                            tot_timehorizon_variable_opex_cost_euro += (
                                results[f"{asset.name}__variable_operational_cost"][0]
                                * optim_time_horizon
                            )
                            tot_timehorizon_fixed_opex_cost_euro += (
                                results[f"{asset.name}__fixed_operational_cost"][0]
                                * optim_time_horizon
                            )

                    except KeyError:
                        # Do not add any costs. Items like joint
                        pass

                # TODO: show discharge energy (current display) and charge energy
                # (new display to be added for ATES etc)
                if (
                    asset.asset_type == "HeatProducer"
                    or asset.asset_type == "GenericProducer"
                    or asset.asset_type == "ResidualHeatSource"
                    or asset.asset_type == "GeothermalSource"
                    or asset.asset_type == "ResidualHeatSource"
                    or asset.asset_type == "GasHeater"
                ):
                    heat_source_energy_wh[asset.name] = np.sum(
                        results[f"{asset.name}.Heat_source"][1:]
                        * (self.times()[1:] - self.times()[0:-1])
                        / 3600
                    )
                # TODO: ATES, HEAT pump show Secondary_heat and Primary_heat and tank storage
                # elif ATES:
                #  summed_charge = np.sum(np.clip(heat_ates, 0.0, np.inf))
                #  summed_discharge = np.abs(np.sum(np.clip(heat_ates, -np.inf, 0.0)))
                # elif Heat pump
                # elif asset.asset_type == "HeatStorage":  # Heat discharged
                #     heat_source_energy_wh[asset.name] = np.sum(
                #         np.clip(results[f"{asset.name}.Heat_buffer"][1:], -np.inf, 0.0)
                #         * (self.times()[1:] - self.times()[0:-1])
                #         / 3600
                #     )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="High level cost breakdown [EUR] (yearly averaged)",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(
                            label="CAPEX",
                            value=(
                                tot_timehorizon_install_cost_euro + tot_timehorizon_invest_cost_euro
                            )
                            / optim_time_horizon,
                        ),
                        esdl.StringItem(
                            label="OPEX",
                            value=tot_variable_opex_cost_euro + tot_fixed_opex_cost_euro,
                        ),
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                ),
            )
        )

        if not optimizer_sim:
            kpis_top_level.kpi.append(
                esdl.DistributionKPI(
                    name=f"High level cost breakdown [EUR] ({optim_time_horizon} year period)",
                    distribution=esdl.StringLabelDistribution(
                        stringItem=[
                            esdl.StringItem(
                                label="CAPEX",
                                value=tot_timehorizon_install_cost_euro
                                + tot_timehorizon_invest_cost_euro,
                            ),
                            esdl.StringItem(
                                label="OPEX",
                                value=(
                                    tot_timehorizon_variable_opex_cost_euro
                                    + tot_timehorizon_fixed_opex_cost_euro
                                ),
                            ),
                        ]
                    ),
                    quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                        physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                    ),
                )
            )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="Overall cost breakdown [EUR] (yearly averaged)",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(
                            label="Installation",
                            value=(tot_timehorizon_install_cost_euro / optim_time_horizon),
                        ),
                        esdl.StringItem(
                            label="Investment",
                            value=(tot_timehorizon_invest_cost_euro / optim_time_horizon),
                        ),
                        esdl.StringItem(label="Variable OPEX", value=tot_variable_opex_cost_euro),
                        esdl.StringItem(label="Fixed OPEX", value=tot_fixed_opex_cost_euro),
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                ),
            )
        )
        if not optimizer_sim:
            kpis_top_level.kpi.append(
                esdl.DistributionKPI(
                    name=f"Overall cost breakdown [EUR] ({optim_time_horizon} year period)",
                    distribution=esdl.StringLabelDistribution(
                        stringItem=[
                            esdl.StringItem(
                                label="Installation", value=tot_timehorizon_install_cost_euro
                            ),
                            esdl.StringItem(
                                label="Investment", value=tot_timehorizon_invest_cost_euro
                            ),
                            esdl.StringItem(
                                label="Variable OPEX",
                                value=tot_timehorizon_variable_opex_cost_euro,
                            ),
                            esdl.StringItem(
                                label="Fixed OPEX",
                                value=tot_timehorizon_fixed_opex_cost_euro,
                            ),
                        ]
                    ),
                    quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                        physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                    ),
                )
            )

            kpis_top_level.kpi.append(
                esdl.DistributionKPI(
                    name=f"CAPEX breakdown [EUR] ({optim_time_horizon} year period)",
                    distribution=esdl.StringLabelDistribution(
                        stringItem=[
                            esdl.StringItem(label=key, value=value)
                            for key, value in asset_timehorizon_capex_breakdown.items()
                        ]
                    ),
                    quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                        physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                    ),
                )
            )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="OPEX breakdown [EUR] (yearly averaged)",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(label=key, value=value)
                        for key, value in asset_opex_breakdown.items()
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                ),
            )
        )
        if not optimizer_sim:
            kpis_top_level.kpi.append(
                esdl.DistributionKPI(
                    name=f"OPEX breakdown [EUR] ({optim_time_horizon} year period)",
                    distribution=esdl.StringLabelDistribution(
                        stringItem=[
                            esdl.StringItem(label=key, value=value)
                            for key, value in asset_timehorizon_opex_breakdown.items()
                        ]
                    ),
                    quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                        physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                    ),
                )
            )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="Energy production [Wh] (yearly averaged)",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(label=key, value=value)
                        for key, value in heat_source_energy_wh.items()
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY, unit=esdl.UnitEnum.WATTHOUR
                ),
            )
        )
        energy_system.instance[0].area.KPIs = kpis_top_level
        # ------------------------------------------------------------------------------------------
        # Cost breakdowns per polygon areas (can consist of several assets of differents types)
        # Notes:
        # - OPEX KPIs are taken into account for energy sources only.
        # - We assume that all energy produced outside of the the subarea comes in via a milp
        #   exchanger that is part of the subarea.
        # TODO: Investigate if no cost in the ESDL then this breaks ESDL visibility
        total_energy_produced_locally_wh = {}
        total_energy_consumed_locally_wh = {}
        estimated_energy_from_local_source_perc = {}
        estimated_energy_from_regional_source_perc = {}

        for subarea in energy_system.instance[0].area.area:
            area_investment_cost = 0.0
            area_installation_cost = 0.0
            area_variable_opex_cost = 0.0
            area_fixed_opex_cost = 0.0

            kpis = esdl.KPIs(id=str(uuid.uuid4()))
            # Here we make a breakdown of the produced energy in the subarea. Where we assume that
            # all energy produced outside of the the subarea comes in via a milp exchanger that is
            # part of the subarea.
            energy_breakdown = {}
            for asset in subarea.asset:
                asset_name = asset.name
                asset_type = self.get_asset_from_asset_name(asset_name).asset_type

                asset_placement_var = self._asset_aggregation_count_var_map[asset.name]
                placed = np.round(results[asset_placement_var][0]) >= 1.0

                if placed:
                    if asset_type == "Joint":
                        continue
                    try:
                        energy_breakdown[asset_type] += np.sum(results[f"{asset_name}.Heat_source"])
                    except KeyError:
                        try:
                            energy_breakdown[asset_type] = np.sum(
                                results[f"{asset_name}.Heat_source"]
                            )
                        except KeyError:
                            try:
                                energy_breakdown[asset_type] += np.sum(
                                    results[f"{asset_name}.Secondary_heat"]
                                )
                            except KeyError:
                                try:
                                    energy_breakdown[asset_type] = np.sum(
                                        results[f"{asset_name}.Secondary_heat"]
                                    )
                                except KeyError:
                                    pass

                    # Create KPIs by using applicable costs for the specific asset
                    area_investment_cost += results[self._asset_investment_cost_map[asset_name]][0]
                    area_installation_cost += results[
                        self._asset_installation_cost_map[asset_name]
                    ][0]
                    area_variable_opex_cost += results[
                        self._asset_variable_operational_cost_map[asset_name]
                    ][0]
                    area_fixed_opex_cost += results[
                        self._asset_fixed_operational_cost_map[asset_name]
                    ][0]

                    # Calculate the total energy [Wh] consumed/produced in an are.
                    # Note: milp losses of buffers, ATES' and pipes are included in the area energy
                    # consumption
                    if asset_name in self.energy_system_components.get("heat_source", []):
                        try:
                            total_energy_produced_locally_wh[subarea.name] += np.sum(
                                results[f"{asset_name}.Heat_source"][1:]
                                * (self.times()[1:] - self.times()[0:-1])
                                / 3600.0
                            )
                        except KeyError:
                            total_energy_produced_locally_wh[subarea.name] = np.sum(
                                results[f"{asset_name}.Heat_source"][1:]
                                * (self.times()[1:] - self.times()[0:-1])
                                / 3600.0
                            )
                    if asset_name in self.energy_system_components.get("heat_demand", []):
                        flow_variable = results[f"{asset_name}.Heat_demand"][1:]
                    elif asset_name in self.energy_system_components.get("heat_buffer", []):
                        flow_variable = results[f"{asset_name}.Heat_buffer"][1:]
                    elif asset_name in self.energy_system_components.get("ates", []):
                        flow_variable = results[f"{asset_name}.Heat_ates"][1:]
                    elif asset_name in self.energy_system_components.get("heat_pipe", []):
                        flow_variable = (
                            np.ones(len(self.times())) * results[f"{asset_name}__hn_heat_loss"]
                        )
                    else:
                        flow_variable = ""
                    if (
                        asset_name in self.energy_system_components.get("heat_demand", [])
                        or asset_name in self.energy_system_components.get("heat_buffer", [])
                        or asset_name in self.energy_system_components.get("ates", [])
                        or asset_name in self.energy_system_components.get("heat_pipe", [])
                    ):
                        try:
                            total_energy_consumed_locally_wh[subarea.name] += np.sum(
                                flow_variable * (self.times()[1:] - self.times()[0:-1]) / 3600.0
                            )
                        except KeyError:
                            total_energy_consumed_locally_wh[subarea.name] = np.sum(
                                flow_variable * (self.times()[1:] - self.times()[0:-1]) / 3600.0
                            )
                    # end Calculate the total energy consumed/produced in an area
                # end if placed loop
            # end asset loop

            # Calculate the estimated energy source [%] for an area
            try:
                if not np.isnan(total_energy_produced_locally_wh[subarea.name]):
                    total_energy_produced_locally_wh_area = total_energy_produced_locally_wh[
                        subarea.name
                    ]
                else:
                    total_energy_produced_locally_wh_area = 0.0
            except KeyError:
                total_energy_produced_locally_wh_area = 0.0

            try:
                if not np.isnan(total_energy_consumed_locally_wh[subarea.name]) and np.isclose(
                    total_energy_consumed_locally_wh[subarea.name], 0.0
                ):
                    estimated_energy_from_local_source_perc[subarea.name] = min(
                        total_energy_produced_locally_wh_area
                        / total_energy_consumed_locally_wh[subarea.name]
                        * 100.0,
                        100.0,
                    )
                    estimated_energy_from_regional_source_perc[subarea.name] = min(
                        (100.0 - estimated_energy_from_local_source_perc[subarea.name]), 100.0
                    )
                else:
                    estimated_energy_from_local_source_perc[subarea.name] = 0.0
                    estimated_energy_from_regional_source_perc[subarea.name] = 0.0
                    logger.warning(
                        f"KPI estimated energy from local & regional source have been set to 0.0"
                        f" in area {subarea.name}. Reason: area local energy"
                        f" consumption value is: {total_energy_consumed_locally_wh[subarea.name]}"
                    )
            except KeyError:
                # Nothing to do, go on to next section of code
                pass

            # Here we add KPIs to the polygon area which allows to visualize them by hovering over
            # it with the mouse
            # Only update kpis if one of the costs > 0, else esdl file will be corrupted
            # TODO: discuss strange behaviour with Edwin - temporarily use of "True" in line below
            if area_investment_cost > 0.0 or area_installation_cost > 0.0 or True:
                kpis.kpi.append(
                    esdl.DoubleKPI(
                        value=area_investment_cost / 1.0e6,
                        name="Investment",
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                            unit=esdl.UnitEnum.EURO,
                            multiplier=esdl.MultiplierEnum.MEGA,
                        ),
                    )
                )
                kpis.kpi.append(
                    esdl.DoubleKPI(
                        value=area_installation_cost / 1.0e6,
                        name="Installation",
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                            unit=esdl.UnitEnum.EURO,
                            multiplier=esdl.MultiplierEnum.MEGA,
                        ),
                    )
                )
            # Only update kpis if one of the costs > 0, else esdl file will be corrupted
            # TODO: discuss strange behaviour with Edwin - temporarily use of "True" in line below
            if area_variable_opex_cost > 0.0 or area_fixed_opex_cost > 0.0 or True:
                kpis.kpi.append(
                    esdl.DoubleKPI(
                        value=area_variable_opex_cost / 1.0e6,
                        name="Variable OPEX (year 1)",
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                            unit=esdl.UnitEnum.EURO,
                            multiplier=esdl.MultiplierEnum.MEGA,
                        ),
                    )
                )
                kpis.kpi.append(
                    esdl.DoubleKPI(
                        value=area_fixed_opex_cost / 1.0e6,
                        name="Fixed OPEX (year 1)",
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                            unit=esdl.UnitEnum.EURO,
                            multiplier=esdl.MultiplierEnum.MEGA,
                        ),
                    )
                )

            try:
                if total_energy_consumed_locally_wh[subarea.name] >= 0.0:
                    kpis.kpi.append(
                        esdl.DoubleKPI(
                            value=round(estimated_energy_from_local_source_perc[subarea.name], 1),
                            name="Estimated energy from local source(s) [%]",
                            quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                                unit=esdl.UnitEnum.PERCENT,
                                multiplier=esdl.MultiplierEnum.NONE,
                            ),
                        )
                    )
                    kpis.kpi.append(
                        esdl.DoubleKPI(
                            value=round(
                                estimated_energy_from_regional_source_perc[subarea.name], 1
                            ),
                            name="Estimated energy from regional source(s) [%]",
                            quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                                unit=esdl.UnitEnum.PERCENT,
                                multiplier=esdl.MultiplierEnum.NONE,
                            ),
                        )
                    )
                    kpis.kpi.append(
                        esdl.DoubleKPI(
                            value=round(total_energy_consumed_locally_wh[subarea.name] / 1.0e9, 1),
                            name="Total energy consumed [GWh]",
                            quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                                physicalQuantity=esdl.PhysicalQuantityEnum.ENERGY,
                                unit=esdl.UnitEnum.WATTHOUR,
                                multiplier=esdl.MultiplierEnum.GIGA,
                            ),
                        )
                    )
            except KeyError:
                # Do nothing because this area does not have any energy consumption
                pass

            # Create plots in the dashboard
            # Top level KPIs: Cost breakdown in a polygon area (for all assest grouped together)
            kpi_name = f"{subarea.name}: Asset cost breakdown [EUR]"
            if (area_installation_cost > 0.0 or area_investment_cost > 0.0) and (
                area_variable_opex_cost > 0.0 or area_fixed_opex_cost > 0.0
            ):
                polygon_area_string_item = [
                    esdl.StringItem(label="Installation", value=area_installation_cost),
                    esdl.StringItem(label="Investment", value=area_investment_cost),
                    esdl.StringItem(label="Variable OPEX", value=area_variable_opex_cost),
                    esdl.StringItem(label="Fixed OPEX", value=area_fixed_opex_cost),
                ]
            elif area_installation_cost > 0.0 or area_investment_cost > 0.0:
                polygon_area_string_item = [
                    esdl.StringItem(label="Installation", value=area_installation_cost),
                    esdl.StringItem(label="Investment", value=area_investment_cost),
                ]
            elif area_variable_opex_cost > 0.0 or area_fixed_opex_cost > 0.0:
                polygon_area_string_item = [
                    esdl.StringItem(label="Variable OPEX", value=area_variable_opex_cost),
                    esdl.StringItem(label="Fixed OPEX", value=area_fixed_opex_cost),
                ]
            if (
                area_installation_cost > 0.0
                or area_investment_cost > 0.0
                or area_variable_opex_cost > 0.0
                or area_fixed_opex_cost > 0.0
            ):
                kpis_top_level.kpi.append(
                    esdl.DistributionKPI(
                        name=kpi_name,
                        distribution=esdl.StringLabelDistribution(
                            stringItem=polygon_area_string_item
                        ),
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                        ),
                    )
                )

            # Here we add a distribution KPI to the subarea to which gives a piechart
            # !!!!!!!!!!!!!!! This will only work if the source is in the area?
            # TODO: Still to be resolved since piecharts are still a work in progress in mapeditor
            # kpis.kpi.append(
            #     esdl.DistributionKPI(
            #         name="Energy breakdown ?",
            #         distribution=esdl.StringLabelDistribution(
            #             stringItem=[
            #                 esdl.StringItem(label=key, value=value) for key,
            #                 value in energy_breakdown.items()
            #             ]
            #         )
            #     )
            # )
            subarea.KPIs = kpis
        # ebd sub-area loop

        # end KPIs

    def _write_updated_esdl(
        self,
        energy_system,
        optimizer_sim: bool = False,
        add_kpis: bool = True,
    ):
        from esdl.esdl_handler import EnergySystemHandler

        results = self.extract_results()
        parameters = self.parameters(0)

        _ = energy_system.id  # input energy system id. Kept here as not sure if still needed
        energy_system.id = str(uuid.uuid4())  # output energy system id
        output_energy_system_id = energy_system.id
        # Currently the simulation_id is created here, but in the future this will probably move
        # to account for 1 simulation/optimization/run potentialy generating more than 1 output
        # energy system (ESDL)
        simulation_id = str(uuid.uuid4())  # simulation (optimization/simulator etc) id

        if optimizer_sim:  # network simulator
            energy_system.name = energy_system.name + "_Simulation"
        else:  # network optimization
            energy_system.name = energy_system.name + "_GrowOptimized"

        def _name_to_asset(name):
            return next(
                (x for x in energy_system.eAllContents() if hasattr(x, "name") and x.name == name)
            )

        if add_kpis:
            self._add_kpis_to_energy_system(energy_system, optimizer_sim)

        # ------------------------------------------------------------------------------------------
        # Placement
        heat_pipes = set(self.energy_system_components.get("heat_pipe", []))
        for _, attributes in self.esdl_assets.items():
            name = attributes.name
            if name in [
                *self.energy_system_components.get("heat_source", []),
                *self.energy_system_components.get("ates", []),
                *self.energy_system_components.get("heat_buffer", []),
            ]:
                asset = _name_to_asset(name)
                asset_placement_var = self._asset_aggregation_count_var_map[name]
                placed = np.round(results[asset_placement_var][0]) >= 1.0
                max_size = results[self._asset_max_size_map[name]][0]

                if asset in self.energy_system_components.get("heat_buffer", []):
                    asset.capacity = max_size
                    asset.volume = max_size / (
                        parameters[f"{name}.cp"]
                        * parameters[f"{name}.rho"]
                        * parameters[f"{name}.dT"]
                    )
                else:
                    asset.power = max_size
                if not placed:
                    asset.delete(recursive=True)
                else:
                    asset.state = esdl.AssetStateEnum.ENABLED
            elif name not in heat_pipes:  # because heat pipes are updated below
                logger.warning(f"ESDL update: asset {name} has not been updated")

        # Pipes:
        edr_pipe_properties_to_copy = ["innerDiameter", "outerDiameter", "diameter", "material"]

        esh_edr = EnergySystemHandler()

        for pipe in self.energy_system_components.get("heat_pipe", []):

            pipe_classes = self.pipe_classes(pipe)
            # When a pipe has not been optimized, enforce pipe to be shown in the simulator
            # ESDL.
            if not pipe_classes:
                if not optimizer_sim:
                    continue
                else:
                    asset.state = esdl.AssetStateEnum.ENABLED

            if not optimizer_sim:
                pipe_class = self.get_optimized_pipe_class(pipe)

            if parameters[f"{pipe}.diameter"] != 0.0 or any(np.abs(results[f"{pipe}.Q"]) > 1.0e-9):
                # if not isinstance(pipe_class, EDRPipeClass):
                #     assert pipe_class.name == f"{pipe}_orig"
                #     continue
                # print(results[f"{pipe}.Q"])
                # print(pipe + " has pipeclass: " + pipe_class.name )
                # print(pipe + f" has diameter: " + pipe_class.name)

                if not optimizer_sim:
                    assert isinstance(pipe_class, EDRPipeClass)
                    asset_edr = esh_edr.load_from_string(pipe_class.xml_string)

                asset = _name_to_asset(pipe)
                asset.state = esdl.AssetStateEnum.ENABLED

                try:
                    asset.costInformation.investmentCosts.value = pipe_class.investment_costs
                except AttributeError:
                    pass
                    # do nothing, in the case that no costs have been specified for the return
                    # pipe in the mapeditor
                except UnboundLocalError:
                    pass

                if not optimizer_sim:
                    for prop in edr_pipe_properties_to_copy:
                        setattr(asset, prop, getattr(asset_edr, prop))
            else:
                asset = _name_to_asset(pipe)
                asset.delete(recursive=True)

        # ------------------------------------------------------------------------------------------
        # Important: This code below must be placed after the "Placement" code. Reason: it relies
        # on unplaced assets being deleted.
        # ------------------------------------------------------------------------------------------
        # Write asset result profile data to database. The database is setup as follows:
        #   - The each time step is represented by a row of data, with columns; datetime, field
        #     values
        #   - The database contains columns based on the carrier connected for visualisation
        #   purposes
        #   - Assets with more than 1 carrier are looped over the time steps as many times as there
        #   are different carriers connected, to ensure the correct data is written to each carrier.
        #   - Database name: input esdl id
        #   - Measurment: carrier id
        #   - Fields: profile value for the specific variable
        #   - Tags used as filters: simulationRun, assetClass, assetName, assetId, capability

        if self.write_result_db_profiles:
            logger.info("Writing asset result profile data to influxDB")
            results = self.extract_results()

            influxdb_conn_settings = ConnectionSettings(
                host=self.influxdb_host,
                port=self.influxdb_port,
                username=self.influxdb_username,
                password=self.influxdb_password,
                database=output_energy_system_id,
                ssl=self.influxdb_ssl,
                verify_ssl=self.influxdb_verify_ssl,
            )

            capabilities = [
                esdl.Transport,
                esdl.Conversion,
                esdl.Consumer,
                esdl.Producer,
            ]

            for asset_name in [
                *self.energy_system_components.get("heat_source", []),
                *self.energy_system_components.get("heat_demand", []),
                *self.energy_system_components.get("heat_pipe", []),
                *self.energy_system_components.get("heat_buffer", []),
                *self.energy_system_components.get("ates", []),
                *self.energy_system_components.get("heat_exchanger", []),
                *self.energy_system_components.get("heat_pump", []),
            ]:
                try:
                    # If the asset has been placed
                    asset = _name_to_asset(asset_name)
                    asset_class = asset.__class__.__name__
                    asset_id = asset.id
                    capability = [c for c in capabilities if c in asset.__class__.__mro__][
                        0
                    ].__name__

                    # Generate three empty variables,
                    # For transport and consumer assets, 'port' is filled with the inport
                    # For producer assets, 'port' is filled with outport as this is linked to the
                    # same carrier as the inport of consumers (thus all info in one carrier)
                    # For conversion assets, the primary side is acting like a consumer, the
                    # secondary side as a producer, thus a similar port structure is assumed, but
                    # now port_prim and port_sec variable are set, such that data can be saved for
                    # both carriers.
                    port, port_prim, port_sec = 3 * [None]
                    if isinstance(asset, esdl.Transport) or isinstance(asset, esdl.Consumer):
                        port = [port for port in asset.port if isinstance(port, esdl.InPort)][0]
                    elif isinstance(asset, esdl.Producer):
                        port = [port for port in asset.port if isinstance(port, esdl.OutPort)][0]
                    elif isinstance(asset, esdl.Conversion):
                        port_prim = [
                            port
                            for port in asset.port
                            if isinstance(port, esdl.InPort) and "Prim" in port.name
                        ][0]
                        port_sec = [
                            port
                            for port in asset.port
                            if isinstance(port, esdl.OutPort) and "Sec" in port.name
                        ][0]
                    else:
                        NotImplementedError(
                            f"influxdb not included for assets of type {type(asset)}"
                        )

                    # Note: when adding new variables to variables_one_hydraulic_system or"
                    # variables_two_hydraulic_system also add quantity and units to the ESDL for
                    # the new variables in the code lower down
                    # These variables exist for all the assets. Variables that only exist for
                    # specific
                    # assets are only added later, like Pump_power
                    variables_one_hydraulic_system = ["HeatIn.Q", "Heat_flow"]
                    variables_two_hydraulic_system = [
                        "Primary.HeatIn.Q",
                        "Secondary.HeatIn.Q",
                        "Heat_flow",
                    ]

                    # Update/overwrite each asset variable list due to:
                    # - the addition of head loss minimization: head variable and pump power
                    # - only a specific variable required for a specific asset: pump power
                    # - addition of post processed variables: pipe velocity
                    if self.heat_network_settings["minimize_head_losses"]:
                        variables_one_hydraulic_system.append("HeatIn.H")
                        variables_two_hydraulic_system.append("Primary.HeatIn.H")
                        variables_two_hydraulic_system.append("Secondary.HeatIn.H")
                        if asset_name in [
                            *self.energy_system_components.get("heat_source", []),
                            *self.energy_system_components.get("heat_buffer", []),
                            *self.energy_system_components.get("ates", []),
                            *self.energy_system_components.get("heat_exchanger", []),
                            *self.energy_system_components.get("heat_pump", []),
                        ]:
                            variables_one_hydraulic_system.append("Pump_power")
                            variables_two_hydraulic_system.append("Pump_power")
                        elif asset_name in [*self.energy_system_components.get("pump", [])]:
                            variables_one_hydraulic_system = ["Pump_power"]
                            variables_two_hydraulic_system = ["Pump_power"]
                    if asset_name in [*self.energy_system_components.get("heat_pipe", [])]:
                        variables_one_hydraulic_system.append("PostProc.Velocity")
                        variables_two_hydraulic_system.append("PostProc.Velocity")
                        # Velocity at the pipe outlet [m/s]
                        post_processed_velocity = (
                            results[f"{asset_name}.HeatOut.Q"] / parameters[f"{asset_name}.area"]
                        )

                    # Depending on the port set, different carriers are assigned
                    if port:
                        carrier_id_dict = {"single_carrier_id": port.carrier.id}
                    elif port_prim and port_sec:
                        carrier_id_dict = {
                            "primary_carrier_id": port_prim.carrier.id,
                            "secondary_carrier_id": port_sec.carrier.id,
                        }
                    else:
                        NotImplementedError(
                            "Unsuported types for the different port carrier combinations"
                        )

                    # Looping over the carrier_ids relevant for the asset
                    # If primary or secondary port are set, variables_to_hydraulic_system will be
                    # used, variable names linking to the secondary port are popped from the list
                    # when the primary port is selected and vice versa
                    variables_two_hydraulic_system_org = variables_two_hydraulic_system.copy()
                    for asset_side, carrier_id in carrier_id_dict.items():
                        variables_two_hydraulic_system = variables_two_hydraulic_system_org.copy()
                        var_pops = []
                        if asset_side == "primary_carrier_id":
                            var_pops = [
                                v for v in variables_two_hydraulic_system if "Secondary" in v
                            ]
                        elif asset_side == "secondary_carrier_id":
                            var_pops = [v for v in variables_two_hydraulic_system if "Primary" in v]
                        for v in var_pops:
                            variables_two_hydraulic_system.remove(v)

                        profiles = ProfileManager()
                        profiles.profile_type = "DATETIME_LIST"
                        profiles.profile_header = ["datetime"]  # + general_headers

                        # Get index of outport which will be used to assign the profile data to
                        index_outport = -1
                        for ip in range(len(asset.port)):
                            if isinstance(asset.port[ip], esdl.OutPort):
                                if index_outport == -1:
                                    index_outport = ip
                                else:
                                    logger.warning(
                                        f"Asset {asset_name} has more than 1 OutPort, and the "
                                        "profile data has been assigned to the 1st OutPort"
                                    )
                                    break

                        if index_outport == -1:
                            logger.error(
                                f"Variable {index_outport} has not been assigned to the asset "
                                f"OutPort"
                            )
                            sys.exit(1)

                        for ii in range(len(self.times())):
                            if not self.io.datetimes[ii].tzinfo:
                                data_row = [
                                    self.io.datetimes[ii].replace(tzinfo=datetime.timezone.utc)
                                ]
                            else:
                                data_row = [self.io.datetimes[ii]]

                            try:
                                # For all components dealing with one hydraulic system
                                if isinstance(
                                    results[f"{asset_name}." + variables_one_hydraulic_system[0]][
                                        ii
                                    ],
                                    numbers.Number,
                                ):
                                    variables_names = variables_one_hydraulic_system
                            except KeyError:
                                # For all components dealing with two hydraulic system
                                if isinstance(
                                    results[f"{asset_name}." + variables_two_hydraulic_system[0]][
                                        ii
                                    ],
                                    numbers.Number,
                                ):
                                    variables_names = variables_two_hydraulic_system
                            except Exception:
                                logger.error(
                                    f"During the influxDB profile writing for asset: {asset_name},"
                                    f" the following error occured:"
                                )
                                traceback.print_exc()
                                sys.exit(1)

                            for variable in variables_names:
                                if ii == 0:
                                    # Set header for each column
                                    profiles.profile_header.append(variable)
                                    # Set profile database attributes for the esdl asset
                                    if not self.io.datetimes[0].tzinfo:
                                        start_date_time = self.io.datetimes[0].replace(
                                            tzinfo=datetime.timezone.utc
                                        )
                                        logger.warning(
                                            f"No timezone specified for the output profile: "
                                            f"default UTC has been used for asset {asset_name} "
                                            f"variable {variable}"
                                        )
                                    else:
                                        start_date_time = self.io.datetimes[0]
                                    if not self.io.datetimes[-1].tzinfo:
                                        end_date_time = self.io.datetimes[-1].replace(
                                            tzinfo=datetime.timezone.utc
                                        )
                                    else:
                                        end_date_time = self.io.datetimes[-1]

                                    profile_attributes = esdl.InfluxDBProfile(
                                        database=output_energy_system_id,
                                        measurement=carrier_id,
                                        field=profiles.profile_header[-1],
                                        port=self.influxdb_port,
                                        host=self.influxdb_host,
                                        startDate=start_date_time,
                                        endDate=end_date_time,
                                        id=str(uuid.uuid4()),
                                        filters='"assetId"=' + f"'{str(asset_id)}'",
                                    )
                                    # Assign quantity and units variable
                                    if variable in ["Heat_flow", "Pump_power"]:
                                        profile_attributes.profileQuantityAndUnit = (
                                            esdl.esdl.QuantityAndUnitType(
                                                physicalQuantity=esdl.PhysicalQuantityEnum.POWER,
                                                unit=esdl.UnitEnum.WATT,
                                                multiplier=esdl.MultiplierEnum.NONE,
                                            )
                                        )
                                    elif variable in [
                                        "HeatIn.H",
                                        "Primary.HeatIn.H",
                                        "Secondary.HeatIn.H",
                                    ]:
                                        profile_attributes.profileQuantityAndUnit = (
                                            esdl.esdl.QuantityAndUnitType(
                                                physicalQuantity=esdl.PhysicalQuantityEnum.PRESSURE,
                                                unit=esdl.UnitEnum.PASCAL,
                                                multiplier=esdl.MultiplierEnum.NONE,
                                            )
                                        )
                                    elif variable in [
                                        "HeatIn.Q",
                                        "Primary.HeatIn.Q",
                                        "Secondary.HeatIn.Q",
                                    ]:
                                        profile_attributes.profileQuantityAndUnit = (
                                            esdl.esdl.QuantityAndUnitType(
                                                physicalQuantity=esdl.PhysicalQuantityEnum.FLOW,
                                                unit=esdl.UnitEnum.CUBIC_METRE,
                                                perTimeUnit=esdl.TimeUnitEnum.SECOND,
                                                multiplier=esdl.MultiplierEnum.NONE,
                                            )
                                        )
                                    elif variable in ["PostProc.Velocity"]:
                                        profile_attributes.profileQuantityAndUnit = (
                                            esdl.esdl.QuantityAndUnitType(
                                                physicalQuantity=esdl.PhysicalQuantityEnum.SPEED,
                                                unit=esdl.UnitEnum.METRE,
                                                perTimeUnit=esdl.TimeUnitEnum.SECOND,
                                                multiplier=esdl.MultiplierEnum.NONE,
                                            )
                                        )
                                    else:
                                        logger.warning(
                                            f"No profile units will be written to the ESDL for: "
                                            f"{asset_name}. + {variable}"
                                        )

                                    asset.port[index_outport].profile.append(profile_attributes)

                                # Add variable values in new column
                                conversion_factor = 0.0
                                if variable in [
                                    "HeatIn.H",
                                    "Primary.HeatIn.H",
                                    "Secondary.HeatIn.H",
                                ]:
                                    conversion_factor = GRAVITATIONAL_CONSTANT * 988.0
                                else:
                                    conversion_factor = 1.0
                                if variable not in ["PostProc.Velocity"]:
                                    data_row.append(
                                        results[f"{asset_name}." + variable][ii] * conversion_factor
                                    )
                                # The variable evaluation below seems unnecessary, but it would be
                                # used we expand the list of post process type variables
                                elif variable in ["PostProc.Velocity"]:
                                    data_row.append(post_processed_velocity[ii])

                            profiles.profile_data_list.append(data_row)
                        # end time steps
                        profiles.num_profile_items = len(profiles.profile_data_list)
                        profiles.start_datetime = profiles.profile_data_list[0][0]
                        profiles.end_datetime = profiles.profile_data_list[-1][0]

                        influxdb_profile_manager = InfluxDBProfileManager(
                            influxdb_conn_settings, profiles
                        )

                        optim_simulation_tag = {
                            "simulationRun": simulation_id,
                            "simulation_type": type(self).__name__,
                            "assetId": asset_id,
                            "assetName": asset_name,
                            "assetClass": asset_class,
                            "capability": capability,
                        }
                        _ = influxdb_profile_manager.save_influxdb(
                            measurement=carrier_id,
                            field_names=influxdb_profile_manager.profile_header[1:],
                            tags=optim_simulation_tag,
                        )

                    # -- Test tags -- # do not delete - to be used in test case
                    # prof_loaded_from_influxdb = InfluxDBProfileManager(influxdb_conn_settings)
                    # dicts = [{"tag": "output_esdl_id", "value": energy_system.id}]
                    # prof_loaded_from_influxdb.load_influxdb(
                    #     # '"' + "ResidualHeatSource_72d7" + '"' ,
                    #     asset_name,
                    #     variables_one_hydraulic_system,
                    #     # ["HeatIn.Q"],
                    #     # ["HeatIn.H"],
                    #     # ["Heat_flow"],
                    #     profiles.start_datetime,
                    #     profiles.end_datetime,
                    #     dicts,
                    # )
                    # test = 0.0

                    # ------------------------------------------------------------------------------
                    # Do not delete the code below: is used in the development of profile viewer in
                    # mapeditor
                    # Write database to excel file and read in to recreate the database
                    # database name: input esdl id
                    # tags when saving to database: optim_simulation_tag = {"output_esdl_id":
                    # output_esdl_id}

                    # print("Save ESDL profile data to excel")
                    # excel_prof_saved = ExcelProfileManager(
                    #     source_profile=prof_loaded_from_influxdb
                    # )
                    # file_path_setting = (
                    #     f"C:\\Projects_gitlab\\NWN_dev\\rtc-tools-milp-network\\{asset_name}.xlsx"
                    # )
                    # excel_prof_saved.save_excel(
                    #     file_path=file_path_setting,
                    #     sheet_name=input_energy_system_id
                    # )
                    # print("Read data from Excel")
                    # excel_prof_read = ExcelProfileManager()
                    # excel_prof_read.load_excel(file_path_setting)
                    # print("Create database")
                    # influxdb_profile_manager_create_new = InfluxDBProfileManager(
                    #     influxdb_conn_settings, excel_prof_read
                    # )
                    # optim_simulation_tag = {"output_esdl_id": energy_system.id}
                    # _ = influxdb_profile_manager_create_new.save_influxdb(
                    #     measurement=asset_name,
                    #     field_names=influxdb_profile_manager_create_new.profile_header[1:],
                    #     tags=optim_simulation_tag,
                    # )
                    # ------------------------------------------------------------------------------
                except StopIteration:
                    # If the asset has been deleted, thus also not placed
                    pass
                except Exception:  # TODO fix other places in the where try/except end with pass
                    logger.error(
                        f"During the influxDB profile writing for asset: {asset_name}, the "
                        "following error occured:"
                    )
                    traceback.print_exc()
                    sys.exit(1)

            # TODO: create test case
            # Code that can be used to remove a specific measurment from the database
            # try:
            #     influxdb_profile_manager.influxdb_client.drop_measurement(energy_system.id)
            # except:
            #     pass
            # Code that can be used to check if a specific measurement exists in the database
            # influxdb_profile_manager.influxdb_client.get_list_measurements()

            # Do not delete: Test code still to be used in test case
            # try:
            #     esdl_infl_prof = profs[0]
            #     np.any(isinstance(esdl_infl_prof, esdl.InfluxDBProfile))
            # except:
            #     np.any(isinstance(profs, esdl.InfluxDBProfile))
            # print("Reading InfluxDB profile from test...")
            # prof3 = InfluxDBProfileManager(conn_settings)
            # # prof3.load_influxdb("test", ["Heat_flow"])
            # prof3.load_influxdb('"' + energy_system.id + '"', profiles.profile_header[1:4])
            # # can access values via
            # # prof3.profile_data_list[0-row][0/1-date/value],
            # # .strftime("%Y-%m-%dT%H:%M:%SZ")
            # # prof3.profile_data_list[3][0].strftime("%Y-%m-%dT%H:%M:%SZ")
            # ts_prof = prof3.get_esdl_timeseries_profile("Heat_flow")
            # # np.testing.assert_array_equal(ts_prof.values[0], 45)
            # # np.testing.assert_array_equal(ts_prof.values[1], 900)
            # # np.testing.assert_array_equal(ts_prof.values[2], 5.6)
            # # np.testing.assert_array_equal(ts_prof.values[3], 1.2)
            # # np.testing.assert_array_equal(len(ts_prof.values), 4)
            # # -- Test tags --
            # prof3 = InfluxDBProfileManager(influxdb_conn_settings)
            # dicts = [{"tag": "output_esdl_id", "value": energy_system.id}]
            # prof3.load_influxdb(
            #     '"' + "ResidualHeatSource_72d7" + '"' , ["HeatIn.Q"],
            #     profiles.start_datetime,
            #     profiles.end_datetime,
            #     dicts,
            # )
            # test = 0.0
        # ------------------------------------------------------------------------------------------
        # Save esdl file

        if self.esdl_parser_class == mesido.esdl.esdl_parser.ESDLFileParser:
            extension = "_Simulation.esdl" if optimizer_sim else "_GrowOptimized.esdl"
            file_path = Path(self.model_folder) / (Path(self.esdl_file_name).stem + extension)
            self.save_energy_system_to_file(energy_system, file_path=file_path)
        self.optimized_esdl_string = self.convert_energy_system_to_string(
            energy_system=energy_system
        )
        # self.__optimized_energy_system_handler = esh
        # self.optimized_esdl_string = esh.to_string()
        #
        # if self.esdl_string is None:
        #     if optimizer_sim:
        #         filename = run_info.esdl_file.with_name(
        #             f"{run_info.esdl_file.stem}_Simulation.esdl"
        #         )
        #     else:
        #         filename = run_info.esdl_file.with_name(
        #             f"{run_info.esdl_file.stem}_GrowOptimized.esdl"
        #         )
        #     esh.save(str(filename))

    def _write_json_output(self):
        # TODO: still add solver stats as json output
        results = self.extract_results()
        workdir = self.output_folder

        parameters = self.parameters(0)
        parameters_dict = dict()
        parameter_path = os.path.join(workdir, "parameters.json")
        for key, value in parameters.items():
            new_value = value  # [x for x in value]
            parameters_dict[key] = new_value
        with open(parameter_path, "w") as file:
            json.dump(parameters_dict, fp=file)

        bounds = self.bounds()
        bounds_dict = dict()
        bounds_path = os.path.join(workdir, "bounds.json")
        for key, value in bounds.items():
            if "Stored_heat" not in key:
                new_value = value  # [x for x in value]
                if isinstance(value[0], Timeseries) or isinstance(value[1], Timeseries):
                    new_value = (value[0].values.tolist(), value[1].values.tolist())
                bounds_dict[key] = new_value
        with open(bounds_path, "w") as file:
            json.dump(bounds_dict, fp=file)

        results_dict = dict()

        for key, values in results.items():
            new_value = values.tolist()
            if len(new_value) == 1:
                new_value = new_value[0]
            results_dict[key] = new_value

        results_path = os.path.join(workdir, "results.json")
        with open(results_path, "w") as file:
            json.dump(results_dict, fp=file)

        # save aliases
        alias_dict = {}
        aliases = self.alias_relation._canonical_variables_map
        for key, values in aliases.items():
            new_value = values
            alias_dict[key] = new_value

        aliases_path = os.path.join(workdir, "aliases.json")
        with open(aliases_path, "w") as file:
            json.dump(alias_dict, fp=file)



import sys
from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.workflows import run_end_scenario_sizing

import numpy as np


class TestUpdatedESDL(TestCase):
    def test_updated_esdl(self):
        """
        Check that the updated ESDL resulting from optmizing a network, is correct by using the
        PoCTutorial and the Grow_workflow. This is done for the actual esdl file and the esdl
        string created by MESIDO. Both these resulting optimized energy systems should be identical
        and it is only the MESIDO esdl saving method that differs.

        Additional checks:
         - Check that the unique profile identification in profile_parser assigns the correct
         profile
         - Check that the correct mulitpler value has been used

        """

        root_folder = str(Path(__file__).resolve().parent.parent)
        sys.path.insert(1, root_folder)

        import examples.PoCTutorial.src.run_grow_tutorial
        from examples.PoCTutorial.src.run_grow_tutorial import EndScenarioSizingStagedHighs

        base_folder = (
            Path(examples.PoCTutorial.src.run_grow_tutorial.__file__).resolve().parent.parent
        )

        problem = run_end_scenario_sizing(
            EndScenarioSizingStagedHighs,
            base_folder=base_folder,
            esdl_file_name="PoC Tutorial.esdl",
            esdl_parser=ESDLFileParser,
        )

        # Save optimized esdl string
        optimized_esdl_string = problem.optimized_esdl_string
        file = open(
            Path.joinpath(base_folder, "model", "PoC Tutorial_GrowOptimized_esdl_string.esdl"), "w"
        )
        file.write(optimized_esdl_string)
        file.close()

        # Check the unique profile identification in profile_parser
        # Test that the correct demand profile is assigned to a demand as expected. Note that the
        # demand profiles are the adapted profiles (peak-hourly, rest-5daily). Therefore the
        # expected max and average hard-coded values are compared to the problem values.
        np.testing.assert_allclose(
            1207800,
            max(problem.get_timeseries("HeatingDemand_b0ff.target_heat_demand").values),
            # demand4_MW, multiplier 0.75, same demand profile as demand HeatingDemand_08fd, but
            # with a different multiplier. So one would expect that this value differs from
            # HeatingDemand_08fd
        )
        np.testing.assert_allclose(
            724680.0,  # demand5_MW, multiplier 0.3
            max(problem.get_timeseries("HeatingDemand_8fbe.target_heat_demand").values),
        )
        np.testing.assert_allclose(
            805200.0,  # demand4_MW, multiplier 0.5
            max(problem.get_timeseries("HeatingDemand_08fd.target_heat_demand").values),
        )
        np.testing.assert_allclose(
            469709.62,  # demand4_MW, multiplier 0.75
            np.average(problem.get_timeseries("HeatingDemand_b0ff.target_heat_demand").values),
        )
        np.testing.assert_allclose(
            281825.77,  # demand5_MW, multiplier 0.3
            np.average(problem.get_timeseries("HeatingDemand_8fbe.target_heat_demand").values),
        )
        np.testing.assert_allclose(
            313139.75,  # demand4_MW, multiplier 0.5
            np.average(problem.get_timeseries("HeatingDemand_08fd.target_heat_demand").values),
        )

        # Checkk that the correct multiplier value was used
        # Compare 2 max values where the same profile was used but with different multiplier values
        # HeatingDemand_08fd: multiplier 0.5
        # HeatingDemand_b0ff: multplier 0.75
        np.testing.assert_allclose(
            max(problem.get_timeseries("HeatingDemand_08fd.target_heat_demand").values) / 0.5,
            max(problem.get_timeseries("HeatingDemand_b0ff.target_heat_demand").values) / 0.75,
        )
        np.testing.assert_allclose(
            np.average(problem.get_timeseries("HeatingDemand_08fd.target_heat_demand").values)
            / 0.5,
            np.average(problem.get_timeseries("HeatingDemand_b0ff.target_heat_demand").values)
            / 0.75,
        )


if __name__ == "__main__":
    import time

    start_time = time.time()

    a = TestUpdatedESDL()
    a.test_updated_esdl()

    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestAtes(TestCase):
    def test_ates(self):
        """
        Checks the constraints concerning the milp to discharge and energy conservation
        for the ates. The heat loss model used are tested and the typical cyclic constraint that
        will be applied in most use cases.

        Checks:
        - the heat loss is computed as expected (loss coef * stored heat [J])
        - checks that the efficiency causes less energy discharged than charged
        - cyclic storage behaviour
        - standard energy conservation, etc.

        """
        import models.test_case_small_network_with_ates.src.run_ates as run_ates
        from models.test_case_small_network_with_ates.src.run_ates import (
            HeatProblem,
        )

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a full year with 365 day timesteps
        solution = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        results = solution.extract_results()

        stored_heat = results["ATES_033c.Stored_heat"]
        heat_loss = results["ATES_033c.Heat_loss"]
        heat_ates = results["ATES_033c.Heat_ates"]
        summed_charge = np.sum(np.clip(heat_ates, 0.0, np.inf))
        summed_discharge = np.abs(np.sum(np.clip(heat_ates, -np.inf, 0.0)))

        # Test if the milp loss is as expected
        coeff = solution.parameters(0)["ATES_033c.heat_loss_coeff"]
        np.testing.assert_allclose(heat_loss, coeff * stored_heat)
        np.testing.assert_array_less(summed_discharge, summed_charge)

        # Test begin and end same state
        np.testing.assert_allclose(stored_heat[0], stored_heat[-1])

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestAtes()
    a.test_ates()
    temp = 0



from .electricity_cable import ElectricityCable
from .electricity_demand import ElectricityDemand
from .electricity_node import ElectricityNode
from .electricity_source import ElectricitySource
from .heat_pump_elec import HeatPumpElec
from .transformer import Transformer

__all__ = [
    "ElectricityNode",
    "ElectricityCable",
    "ElectricitySource",
    "ElectricityDemand",
    "HeatPumpElec",
    "Transformer",
]



import base64
import copy
import logging
import xml.etree.ElementTree as ET  # noqa: N817
from datetime import timedelta
from pathlib import Path
from typing import Any, Dict, Optional

import esdl.esdl_handler

from mesido.component_type_mixin import (
    ModelicaComponentTypeMixin,
)
from mesido.esdl.asset_to_component_base import _AssetToComponentBase
from mesido.esdl.common import Asset
from mesido.esdl.edr_pipe_class import EDRGasPipeClass, EDRPipeClass
from mesido.esdl.esdl_heat_model import ESDLHeatModel
from mesido.esdl.esdl_model_base import _ESDLModelBase
from mesido.esdl.esdl_parser import ESDLStringParser
from mesido.esdl.esdl_qth_model import ESDLQTHModel
from mesido.esdl.profile_parser import BaseProfileReader, InfluxDBProfileReader
from mesido.physics_mixin import PhysicsMixin
from mesido.pipe_class import GasPipeClass, PipeClass
from mesido.pycml.pycml_mixin import PyCMLMixin
from mesido.qth_not_maintained.qth_mixin import QTHMixin

import numpy as np

import rtctools.data.pi as pi
from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.optimization.io_mixin import IOMixin


logger = logging.getLogger("mesido")


ns = {"fews": "http://www.wldelft.nl/fews", "pi": "http://www.wldelft.nl/fews/PI"}
DEFAULT_START_TIMESTAMP = "2017-01-01T00:00:00+00:00"
DEFAULT_END_TIMESTAMP = "2018-01-01T00:00:00+00:00"


class _ESDLInputException(Exception):
    pass


class ESDLMixin(
    ModelicaComponentTypeMixin,
    IOMixin,
    PyCMLMixin,
    CollocatedIntegratedOptimizationProblem,
):
    """
    This class is used to be able to parse an ESDL file and utilize the definition of the energy
    system in that file. Furthermore, it contains functionality to extract profiles specified like
    for example demand profiles.
    """

    esdl_run_info_path: Path = None

    esdl_pi_validate_timeseries: bool = False

    __max_supply_temperature: Optional[float] = None

    # TODO: remove this once ESDL allows specifying a minimum pipe size for an optional pipe.
    __minimum_pipe_size_name: str = "DN150"

    def __init__(self, *args, **kwargs) -> None:
        """
        In this __init__ function we do the parsing of the esdl file based on either a string which
        is provided or read it in from a provided file name.

        We put the assets, profiles and carriers in attributes of the class to later instantiate
        the PyCML objects and write the desired time-series.

        We set file locations for the input files and for the diagnostic file.

        We create a dict with all possible pipe classes for the optional pipes to later add them
        to the optimization problem. This is done in this Mixin as we here use the information of
        the EDR database which is linked to ESDL and the Mapeditor.

        Parameters
        ----------
        args : none
        kwargs : esdl_string or esdl_file_name must be provided
        """

        self.esdl_parser_class: type = kwargs.get("esdl_parser", ESDLStringParser)
        esdl_string = kwargs.get("esdl_string", None)
        model_folder = kwargs.get("model_folder")
        esdl_file_name = kwargs.get("esdl_file_name", None)
        esdl_path = None
        if esdl_file_name is not None:
            esdl_path = Path(model_folder) / esdl_file_name

        # TODO: discuss if this is correctly located here and why the reading of profiles is then
        #  in the read function?
        esdl_parser = self.esdl_parser_class(esdl_string=esdl_string, esdl_path=esdl_path)
        esdl_parser.read_esdl()
        self._esdl_assets: Dict[str, Asset] = esdl_parser.get_assets()
        self._esdl_carriers: Dict[str, Dict[str, Any]] = esdl_parser.get_carrier_properties()
        self.__energy_system_handler: esdl.esdl_handler.EnergySystemHandler = esdl_parser.get_esh()

        profile_reader_class = kwargs.get("profile_reader", InfluxDBProfileReader)
        input_file_name = kwargs.get("input_timeseries_file", None)
        input_folder = kwargs.get("input_folder")
        input_file_path = None
        if input_file_name is not None:
            input_file_path = Path(input_folder) / input_file_name
        self.__profile_reader: BaseProfileReader = profile_reader_class(
            energy_system=self.__energy_system_handler.energy_system, file_path=input_file_path
        )

        # This way we allow users to adjust the parsed ESDL assets
        assets = self.esdl_assets

        # Although we work with the names, the FEWS import data uses the component IDs
        self.__timeseries_id_map = {a.id: a.name for a in assets.values()}

        if isinstance(self, PhysicsMixin):
            self.__model = ESDLHeatModel(assets, **self.esdl_heat_model_options())
        else:
            assert isinstance(self, QTHMixin)

            # Maximum supply temperature is very network dependent, so it is
            # hard to choose a default. Therefore, we look at the global
            # properties instead and add 10 degrees on top.
            global_supply_temperatures = [
                c["temperature"]
                for a in assets.values()
                for c in a.global_properties["carriers"].values()
            ]
            max_global_supply = max(x for x in global_supply_temperatures if np.isfinite(x))

            attribute_temperatures = [
                a.attributes.get("maxTemperature", -np.inf) for a in assets.values()
            ]
            max_attribute = max(x for x in attribute_temperatures if np.isfinite(x))

            self.__max_supply_temperature = max(max_global_supply, max_attribute) + 10.0

            self.__model = ESDLQTHModel(assets, **self.esdl_qth_model_options())

        self._override_pipe_classes = dict()
        self.override_pipe_classes()
        self._override_gas_pipe_classes = dict()
        self.override_gas_pipe_classes()

        self.name_to_esdl_id_map = dict()

        super().__init__(*args, **kwargs)

    @property
    def esdl_bytes_string(self) -> bytes:
        """
        Returns a bytes string representation of the ESDL model used.
        """
        return base64.b64encode(self.__energy_system_handler.to_string().encode("utf-8"))

    def pre(self) -> None:
        """
        In this pre method we create a dict with a mapping between the esdl id and the name. We
        also check that every asset has an unique name, which is needed for us to create unique
        variable names.

        Returns
        -------
        None
        """
        super().pre()
        for esdl_id, esdl_asset in self.esdl_assets.items():
            if esdl_asset.name in self.name_to_esdl_id_map:
                raise RuntimeWarning(
                    f"Found multiple ESDL assets with name {esdl_asset.name} in the "
                    f"input ESDL. This is not supported in the optimization."
                )
            self.name_to_esdl_id_map[esdl_asset.name] = esdl_id

    def override_pipe_classes(self) -> None:
        """
        In this method we populate the _override_pipe_classes dict, which gives a list of possible
        pipe classes for every pipe. We do this only when a pipe has the state OPTIONAL. We use the
        EDR pipe classes. We assume that it is possible to remove a pipe PipeClass None, but also
        that there is a minimum layed pipe size of DN150 to limit the search space. This seems
        reasonable as we focus upon regional and primary networks.

        Returns
        -------
        None
        """
        maximum_velocity = self.heat_network_settings["maximum_velocity"]

        no_pipe_class = PipeClass("None", 0.0, 0.0, (0.0, 0.0), 0.0)
        pipe_classes = [
            EDRPipeClass.from_edr_class(name, edr_class_name, maximum_velocity)
            for name, edr_class_name in _AssetToComponentBase.STEEL_S1_PIPE_EDR_ASSETS.items()
        ]

        # We assert the pipe classes are monotonically increasing in size
        assert np.all(np.diff([pc.inner_diameter for pc in pipe_classes]) > 0)

        for asset in self.esdl_assets.values():
            if asset.asset_type == "Pipe" and isinstance(
                asset.in_ports[0].carrier, esdl.HeatCommodity
            ):
                p = asset.name

                if asset.attributes["state"].name == "OPTIONAL":
                    c = self._override_pipe_classes[p] = []
                    c.append(no_pipe_class)

                    min_size = self.__minimum_pipe_size_name
                    min_size_idx = [
                        idx for idx, pipe in enumerate(pipe_classes) if pipe.name == min_size
                    ]
                    assert len(min_size_idx) == 1
                    min_size_idx = min_size_idx[0]

                    max_size = asset.attributes["diameter"].name

                    max_size_idx = [
                        idx for idx, pipe in enumerate(pipe_classes) if pipe.name == max_size
                    ]
                    assert len(max_size_idx) == 1
                    max_size_idx = max_size_idx[0]

                    if max_size_idx < min_size_idx:
                        logger.warning(
                            f"{p} has an upper DN size smaller than the used minimum size "
                            f"of {self.__minimum_pipe_size_name}, choose at least "
                            f"{self.__minimum_pipe_size_name}"
                        )
                    elif min_size_idx == max_size_idx:
                        c.append(pipe_classes[min_size_idx])
                    else:
                        c.extend(pipe_classes[min_size_idx : max_size_idx + 1])
                elif asset.attributes["state"].name == "DISABLED":
                    c = self._override_pipe_classes[p] = []
                    c.append(no_pipe_class)

    def override_gas_pipe_classes(self) -> None:
        """
        In this method we populate the _override_gas_pipe_classes dict, which gives a list of
        possible pipe classes for every pipe. We do this only when a pipe has the state OPTIONAL.
        We use the EDR pipe classes. We assume that it is possible to remove a pipe PipeClass None,
        but also that there is a minimum layed pipe size of DN150 to limit the search space. This
        seems reasonable as we focus upon regional and primary networks.

        Returns
        -------
        None
        """
        maximum_velocity = self.gas_network_settings["maximum_velocity"]

        no_pipe_class = GasPipeClass("None", 0.0, 0.0, 0.0)
        pipe_classes = [
            EDRGasPipeClass.from_edr_class(name, edr_class_name, maximum_velocity)
            for name, edr_class_name in _AssetToComponentBase.STEEL_S1_PIPE_EDR_ASSETS.items()
        ]

        # We assert the pipe classes are monotonically increasing in size
        assert np.all(np.diff([pc.inner_diameter for pc in pipe_classes]) > 0)

        for asset in self.esdl_assets.values():
            if asset.asset_type == "Pipe" and isinstance(
                asset.in_ports[0].carrier, esdl.GasCommodity
            ):
                p = asset.name

                if asset.attributes["state"].name == "OPTIONAL":
                    c = self._override_gas_pipe_classes[p] = []
                    c.append(no_pipe_class)

                    min_size = self.__minimum_pipe_size_name
                    min_size_idx = [
                        idx for idx, pipe in enumerate(pipe_classes) if pipe.name == min_size
                    ]
                    assert len(min_size_idx) == 1
                    min_size_idx = min_size_idx[0]

                    max_size = asset.attributes["diameter"].name

                    max_size_idx = [
                        idx for idx, pipe in enumerate(pipe_classes) if pipe.name == max_size
                    ]
                    assert len(max_size_idx) == 1
                    max_size_idx = max_size_idx[0]

                    if max_size_idx < min_size_idx:
                        logger.warning(
                            f"{p} has an upper DN size smaller than the used minimum size "
                            f"of {self.__minimum_pipe_size_name}, choose at least "
                            f"{self.__minimum_pipe_size_name}"
                        )
                    elif min_size_idx == max_size_idx:
                        c.append(pipe_classes[min_size_idx])
                    else:
                        c.extend(pipe_classes[min_size_idx : max_size_idx + 1])
                elif asset.attributes["state"].name == "DISABLED":
                    c = self._override_gas_pipe_classes[p] = []
                    c.append(no_pipe_class)

    @property
    def esdl_assets(self) -> Dict[str, Asset]:
        """
        property method to retrieve the esdl assets which are a private attribute of the class.

        Returns
        -------
        A dict of the esdl assets with their properties
        """
        return self._esdl_assets

    @property
    def esdl_carriers(self, type=None) -> Dict[str, Dict[str, Any]]:
        """
        property method to retrieve the esdl carriers which are a private attribute of the class.

        Returns
        -------
        A dict with the id of the carrier and the attributes in the value
        """

        return self._esdl_carriers

    def esdl_carriers_typed(self, type=None) -> Dict[str, Dict[str, Any]]:
        """
        property method to retrieve the esdl carriers which are a private attribute of the class.

        Returns
        -------
        A dict with the id of the carrier and the attributes in the value
        """
        if type is None:
            return self._esdl_carriers
        else:
            carriers = {}
            for id, attr in self._esdl_carriers.items():
                if attr["type"] in type:
                    carriers[id] = attr
        return carriers

    def get_energy_system_copy(self) -> esdl.esdl.EnergySystem:
        """
        Method to get a copy of the energy system loaded that can be edited without touching the
        original

        Returns
        -------
        A copy of the energy system loaded
        """
        return copy.deepcopy(self.__energy_system_handler.energy_system)

    @staticmethod
    def convert_energy_system_to_string(energy_system: esdl.esdl.EnergySystem) -> str:
        """
        Method to convert a given energy system into a string using a copy of the energy system
        handler that is available within this class

        Returns
        -------
        An XML string representing the energy system
        """

        uri = esdl.esdl_handler.StringURI("to_string.esdl")
        energy_system.eResource.save(uri)
        return uri.getvalue()

    @staticmethod
    def save_energy_system_to_file(energy_system: esdl.esdl.EnergySystem, file_path: Path) -> None:
        """
        Method to save a given energy system to file (using the standard ESDL XML schema, using the
        energy system handler available within this class

        Returns
        -------
        None
        """
        esh = esdl.esdl_handler.EnergySystemHandler(energy_system=energy_system)
        esh.save(filename=str(file_path))

    @property
    def esdl_asset_id_to_name_map(self) -> Dict:
        """
        A map between the id and the name of an asset. Very bad naming of the attribute...

        Returns
        -------
        A dict with the id to name map.
        """
        return self.__timeseries_id_map.copy()

    @property
    def esdl_asset_name_to_id_map(self) -> Dict:
        """
        A map between the name and the id of an asset. Very bad naming of the attribute...

        Returns
        -------
        A dict with the name to id map.
        """
        return dict(zip(self.__timeseries_id_map.values(), self.__timeseries_id_map.keys()))

    def get_asset_from_asset_name(self, asset_name: str) -> esdl.Asset:
        """
        This function returns the esdl asset with its properties based on the name you provide

        Parameters
        ----------
        asset_name : string with the asset name of the esdl asset.

        Returns
        -------
        The esdl asset with its attributes and global properties
        """

        asset_id = self.esdl_asset_name_to_id_map[asset_name]
        return self.esdl_assets[asset_id]

    def esdl_heat_model_options(self) -> Dict:
        """
        function to spedifically return the needed HeatMixin options needed for the conversion
        from ESDL to pycml. This case velocities used to set nominals and caps on the milp.

        Returns
        -------
        dict with estimated and maximum velocity
        """
        energy_system_options = self.energy_system_options()
        v_nominal = energy_system_options["estimated_velocity"]
        v_max = self.heat_network_settings["maximum_velocity"]
        return dict(v_nominal=v_nominal, v_max=v_max)

    def esdl_qth_model_options(self) -> Dict:
        """
        function to spedifically return the needed HeatMixin options needed for the conversion
        from ESDL to pycml. This case velocities used to set nominals and caps on the milp.

        Returns
        -------
        dict with estimated and maximum velocity
        """
        heat_network_options = self.energy_system_options()
        kwargs = {}
        kwargs["v_nominal"] = heat_network_options["estimated_velocity"]
        kwargs["v_max"] = self.heat_network_settings["maximum_velocity"]
        if self.__max_supply_temperature is not None:
            kwargs["maximum_temperature"] = self.__max_supply_temperature
        return dict(**kwargs)

    def is_hot_pipe(self, pipe: str) -> bool:
        """
        To check if a pipe is part of the "supply" network.

        Parameters
        ----------
        pipe : string with name of the pipe

        Returns
        -------
        Returns true if the pipe is in the supply network thus not ends with "_ret"
        """
        return True if pipe not in self.cold_pipes else False

    def is_cold_pipe(self, pipe: str) -> bool:
        """
        To check if a pipe is part of the "return" network. Note we only assign to the return
        network if it has a dedicated hot pipe.

        Parameters
        ----------
        pipe : string with name of the pipe

        Returns
        -------
        Returns true if the pipe is in the return network thus ends with "_ret"
        """
        return pipe.endswith("_ret")

    def hot_to_cold_pipe(self, pipe: str) -> str:
        """
        To get the name of the respective cold pipe. Note hot pipes do not automatically have a
        dedicated return pipe in case of different supply and return topologies and/or temperature
        cascading. This function should only be called if the cold pipe exists.

        Parameters
        ----------
        pipe : string with hot pipe name.

        Returns
        -------
        string with the associated return pipe name.
        """
        return f"{pipe}_ret"

    def cold_to_hot_pipe(self, pipe: str) -> str:
        """
        To get the name of the respective hot pipe. Note hot pipes do not automatically have a
        dedicated return pipe in case of different supply and return topologies and/or temperature
        cascading.

        Parameters
        ----------
        pipe : string with cold pipe name.

        Returns
        -------
        string with the associated hot pipe name.
        """
        return pipe[:-4]

    def pycml_model(self) -> _ESDLModelBase:
        """
        Function to get the model description.

        Returns
        -------
        Returns the pycml model object.
        """
        return self.__model

    def read(self) -> None:
        """
        In this read function we read the relevant time-series and write them to the io object for
        later use. We read and write the demand and production profiles. These profiles can either
        be specified in the esdl file referring to an InfluxDB profile, or be specified in a csv
        file in this case we rely on the user to give the csv file in the runinfo.xml.

        Returns
        -------
        None
        """
        super().read()
        energy_system_components = self.energy_system_components
        esdl_carriers = self.esdl_carriers
        io = self.io
        self.__profile_reader.read_profiles(
            energy_system_components=energy_system_components,
            io=io,
            esdl_asset_id_to_name_map=self.esdl_asset_id_to_name_map,
            esdl_assets=self.esdl_assets,
            carrier_properties=esdl_carriers,
            ensemble_size=self.ensemble_size,
        )

    def write(self) -> None:
        """
        This function comes from legacy with CF in the WarmingUP time. It was used to write out a
        xml file with in that the timeseries output of some specified types of assets. This method
        works but is no longer maintained.

        Returns
        -------
        None
        """
        super().write()

        if getattr(self, "__output_timeseries_file", None) is None:
            return

        output_timeseries_file = Path(self.__output_timeseries_file)
        assert output_timeseries_file.is_absolute()
        assert output_timeseries_file.suffix == ".xml"

        timeseries_export_basename = output_timeseries_file.stem
        output_folder = output_timeseries_file.parent

        try:
            self.__timeseries_export = pi.Timeseries(
                self.esdl_pi_output_data_config(self.__timeseries_id_map),
                output_folder,
                timeseries_export_basename,
                binary=False,
                pi_validate_times=self.esdl_pi_validate_timeseries,
            )
        except IOError:
            raise Exception(
                "ESDLMixin: {}.xml not found in {}.".format(
                    timeseries_export_basename, output_folder
                )
            )

        # Get time stamps
        times = self.times()
        if len(set(times[1:] - times[:-1])) == 1:
            dt = timedelta(seconds=times[1] - times[0])
        else:
            dt = None

        output_keys = [k for k, v in self.__timeseries_export.items()]

        # Start of write output
        # Write the time range for the export file.
        self.__timeseries_export.times = [
            self.__timeseries_import.times[self.__timeseries_import.forecast_index]
            + timedelta(seconds=s)
            for s in times
        ]

        # Write other time settings
        self.__timeseries_export.forecast_datetime = self.__timeseries_import.forecast_datetime
        self.__timeseries_export.dt = dt
        self.__timeseries_export.timezone = self.__timeseries_import.timezone

        # Write the ensemble properties for the export file.
        if self.ensemble_size > 1:
            self.__timeseries_export.contains_ensemble = True
        self.__timeseries_export.ensemble_size = self.ensemble_size
        self.__timeseries_export.contains_ensemble = self.ensemble_size > 1

        # Start looping over the ensembles for extraction of the output values.
        for ensemble_member in range(self.ensemble_size):
            results = self.extract_results(ensemble_member)

            # For all variables that are output variables the values are
            # extracted from the results.
            for variable in output_keys:
                try:
                    values = results[variable]
                    if len(values) != len(times):
                        values = self.interpolate(
                            times,
                            self.times(variable),
                            values,
                            self.interpolation_method(variable),
                        )
                except KeyError:
                    try:
                        ts = self.get_timeseries(variable, ensemble_member)
                        if len(ts.times) != len(times):
                            values = self.interpolate(times, ts.times, ts.values)
                        else:
                            values = ts.values
                    except KeyError:
                        logger.warning(
                            "ESDLMixin: Output requested for non-existent variable {}. "
                            "Will not be in output file.".format(variable)
                        )
                        continue

                self.__timeseries_export.set(variable, values, ensemble_member=ensemble_member)

        # Write output file to disk
        self.__timeseries_export.write()


class _ESDLInputDataConfig:
    """
    This class is used to specify naming standard for input data, specifically for demand and
    production profiles.
    """

    def __init__(self, id_map: Dict, energy_system_components: Dict) -> None:
        # TODO: change naming source and demand to heat_source and heat_demand throughout code
        self.__id_map = id_map
        self._sources = set(energy_system_components.get("heat_source", []))
        self._demands = set(energy_system_components.get("heat_demand", []))
        self._electricity_sources = set(energy_system_components.get("electricity_source", []))
        self._electricity_demands = set(energy_system_components.get("electricity_demand", []))
        self._gas_sources = set(energy_system_components.get("gas_source", []))
        self._gas_demands = set(energy_system_components.get("gas_demand", []))

    def variable(self, pi_header: Any) -> str:
        """
        Old function not maintained anymore from WarmingUp times. The input xml file would specify
        the id of the asset for which a time-series was given. This function would return the name
        we use in our framework for that same variable.

        Parameters
        ----------
        pi_header : the xml header element in which the id of the asset is specified

        Returns
        -------
        string with the name of the timeseries name.
        """
        location_id = pi_header.find("pi:locationId", ns).text

        try:
            component_name = self.__id_map[location_id]
        except KeyError:
            parameter_id = pi_header.find("pi:parameterId", ns).text
            qualifiers = pi_header.findall("pi:qualifierId", ns)
            qualifier_ids = ":".join(q.text for q in qualifiers)
            return f"{location_id}:{parameter_id}:{qualifier_ids}"

        if component_name in self._demands:
            suffix = ".target_heat_demand"
        elif component_name in self._sources:
            suffix = ".maximum_heat_source"
        elif component_name in self._electricity_demands:
            suffix = ".target_electricity_demand"
        elif component_name in self._electricity_sources:
            suffix = ".maximum_electricity_source"
        elif component_name in self._gas_demands:
            suffix = ".target_gas_demand"
        elif component_name in self._gas_sources:
            suffix = ".maximum_gas_source"
        else:
            logger.warning(
                f"Could not identify '{component_name}' as either source or demand. "
                f"Using neutral suffix '.target_heat' for its milp timeseries."
            )
            suffix = ".target_heat"

        # Note that the qualifier id (if any specified) refers to the profile
        # element of the respective ESDL asset->in_port. For now we just
        # assume that only milp demand timeseries are set in the XML file.
        return f"{component_name}{suffix}"

    def pi_variable_ids(self, variable):
        raise NotImplementedError

    def parameter(self, parameter_id, location_id=None, model_id=None):
        raise NotImplementedError

    def pi_parameter_ids(self, parameter):
        raise NotImplementedError


class _ESDLOutputDataConfig:
    def __init__(self, id_map):
        self.__id_map = id_map

    def variable(self, pi_header):
        location_id = pi_header.find("pi:locationId", ns).text
        parameter_id = pi_header.find("pi:parameterId", ns).text

        component_name = self.__id_map[location_id]

        return f"{component_name}.{parameter_id}"

    def pi_variable_ids(self, variable):
        raise NotImplementedError

    def parameter(self, parameter_id, location_id=None, model_id=None):
        raise NotImplementedError

    def pi_parameter_ids(self, parameter):
        raise NotImplementedError


def _overwrite_parameters(parameters_file, assets):
    paramroot = ET.parse(parameters_file).getroot()
    groups = paramroot.findall("pi:group", ns)

    for parameter in groups:
        id_ = parameter.attrib["id"]
        param_name = parameter[0].attrib["id"]
        param_value = parameter[0][0].text

        asset = assets[id_]
        type_ = type(asset.attributes[param_name])
        asset.attributes[param_name] = type_(param_value)

    return assets



from abc import abstractmethod
from typing import Dict, List

from .topology import Topology


class BaseComponentTypeMixin:
    """
    The standard naming convention is that pipes have "_hot" and "_cold" suffixes.
    Such convention can be overridden using the `is_hot_pipe` and `is_cold_pipe` methods.
    Moreover, one has to set the mapping between hot and cold pipes via `hot_to_cold_pipe`
    and `cold_to_hot_pipe`.
    """

    @property
    @abstractmethod
    def energy_system_components(self) -> Dict[str, str]:
        """
        This method return a dict with the components structured by asset_type.
        """
        raise NotImplementedError

    def energy_system_components_get(self, list_types: list) -> list:
        components = []
        for component_type in list_types:
            components.extend(self.energy_system_components.get(component_type, []))
        components = list(set(components))
        return components

    @property
    @abstractmethod
    def energy_system_topology(self) -> Topology:
        """
        The method return a Topology object that contains asset with specific
        topology and associated direction information needed for the physics in the constraints
        """
        raise NotImplementedError

    def is_hot_pipe(self, pipe: str) -> bool:
        """
        The function return true if the pipe is a supply pipe based on a name convention
        """
        return pipe.endswith("_hot")

    def is_cold_pipe(self, pipe: str) -> bool:
        """
        The function return true if the pipe is a return pipe based on a name convention
        """
        return pipe.endswith("_cold")

    def hot_to_cold_pipe(self, pipe: str):
        """
        This function returns the name of the associated cold/return pipe of a supply/hot pipe.
        """
        assert self.is_hot_pipe(pipe)

        return f"{pipe[:-4]}_cold"

    def cold_to_hot_pipe(self, pipe: str):
        """
        This function returns the name of the associated hot/supply pipe of a cold/return pipe.
        """
        assert self.is_cold_pipe(pipe)

        return f"{pipe[:-5]}_hot"

    def has_related_pipe(self, pipe: str) -> bool:
        """
        This function checks whether a pipe has a related hot/cold pipe. This is done based on the
        name convention.

        :params pipe: is the pipe name.
        :returns: True if the pipe has a related pipe, else False.
        """
        related = False
        if self.is_hot_pipe(pipe):
            if self.hot_to_cold_pipe(pipe) in self.energy_system_components.get("heat_pipe", []):
                related = True
        elif self.is_cold_pipe(pipe):
            if self.cold_to_hot_pipe(pipe) in self.energy_system_components.get("heat_pipe", []):
                related = True
        return related

    @property
    def hot_pipes(self) -> List[str]:
        """
        This function return a list of all the supply/hot pipe names.
        """
        return [
            p for p in self.energy_system_components.get("heat_pipe", []) if self.is_hot_pipe(p)
        ]

    @property
    def cold_pipes(self) -> List[str]:
        """
        This function return a list of all the return/cold pipe names.
        """
        return [
            p for p in self.energy_system_components.get("heat_pipe", []) if self.is_cold_pipe(p)
        ]



import logging
from typing import Dict, Set

from mesido.base_component_type_mixin import BaseComponentTypeMixin
from mesido.heat_network_common import NodeConnectionDirection
from mesido.topology import Topology

from pymoca.backends.casadi.alias_relation import AliasRelation

logger = logging.getLogger("mesido")


class ModelicaComponentTypeMixin(BaseComponentTypeMixin):
    """
    This class is used to make the milp network component information easily accesible in the
    heat_mixin. This is achieved by creating a heat_network_components dict where the assets are
    sorted by asset_type. Furthermore, the topology object is created in which for specific assets
    the connections with directions are saved for later use in the constraints.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        self.__hn_component_types = None

    def pre(self):
        """
        In this function the topology object of the milp network is constructed. Meaning that for
        nodes, busses and storage assets their relevant information on the positive flow direction
        and connections on the ports is gathered and stored in the topology object.
        """
        components = self.energy_system_components
        nodes = components.get("node", [])
        busses = components.get("electricity_node", [])
        gas_nodes = components.get("gas_node", [])
        buffers = components.get("heat_buffer", [])
        atess = [*components.get("ates", []), *components.get("low_temperature_ates", [])]
        try:
            pipes = components["heat_pipe"]
            cables = components.get("electricity_cable", [])
            gas_pipes = components.get("gas_pipe", [])
        except KeyError:
            try:
                cables = components["electricity_cable"]
                gas_pipes = components.get("gas_pipe", [])
                pipes = []
            except KeyError:
                try:
                    cables = []
                    gas_pipes = components["gas_pipe"]
                    pipes = []
                except KeyError:
                    logger.error(
                        "A valid network should have at least one pipe/cable, "
                        "assets cannot be connected directly"
                    )

        # Figure out which pipes are connected to which nodes, which pipes
        # are connected in series, and which pipes are connected to which buffers.

        pipes_set = set(pipes)
        cables_set = set(cables)
        gas_pipes_set = set(gas_pipes)
        parameters = [self.parameters(e) for e in range(self.ensemble_size)]
        node_connections = {}
        bus_connections = {}
        gas_node_connections = {}

        # Figure out if we are dealing with a Heat model, or a QTH model
        try:
            if len(pipes):
                _ = self.variable(f"{pipes[0]}.HeatIn.Heat")
                heat_network_model_type = "Heat"
            else:
                heat_network_model_type = "Heat"
        except KeyError:
            heat_network_model_type = "QTH"

        for n in [*nodes, *busses, *gas_nodes]:
            n_connections = [ens_params[f"{n}.n"] for ens_params in parameters]

            if len(set(n_connections)) > 1:
                raise Exception(
                    "Nodes and busses cannot have differing number of connections per "
                    "ensemble member"
                )

            n_connections = n_connections[0]

            # Note that we do this based on temperature, because discharge may
            # be an alias of yet some other further away connected pipe.
            if n in nodes:
                node_connections[n] = connected_pipes = {}
            elif n in busses:
                bus_connections[n] = connected_pipes = {}
            elif n in gas_nodes:
                gas_node_connections[n] = connected_pipes = {}

            for i in range(n_connections):
                if n in nodes:
                    cur_port = f"{n}.{heat_network_model_type}Conn[{i + 1}]"
                    prop = "T" if heat_network_model_type == "QTH" else "Heat"
                    in_suffix = ".QTHIn.T" if heat_network_model_type == "QTH" else ".HeatIn.Heat"
                    out_suffix = (
                        ".QTHOut.T" if heat_network_model_type == "QTH" else ".HeatOut.Heat"
                    )
                elif n in busses:
                    cur_port = f"{n}.ElectricityConn[{i + 1}]"
                    prop = "Power"
                    in_suffix = ".ElectricityIn.Power"
                    out_suffix = ".ElectricityOut.Power"
                elif n in gas_nodes:
                    # TODO: Ideally a temporary variable would be created to make the connections
                    #  map that is not passed to the problem
                    cur_port = f"{n}.GasConn[{i + 1}]"
                    prop = "Q_shadow"
                    in_suffix = ".GasIn.Q_shadow"
                    out_suffix = ".GasOut.Q_shadow"
                aliases = [
                    x
                    for x in self.alias_relation.aliases(f"{cur_port}.{prop}")
                    if not x.startswith(n) and x.endswith(f".{prop}")
                ]

                if len(aliases) > 1:
                    raise Exception(f"More than one connection to {cur_port}")
                elif len(aliases) == 0:
                    raise Exception(f"Found no connection to {cur_port}")

                if aliases[0].endswith(out_suffix):
                    pipe_w_orientation = (
                        aliases[0][: -len(out_suffix)],
                        NodeConnectionDirection.IN,
                    )
                else:
                    assert aliases[0].endswith(in_suffix)
                    pipe_w_orientation = (
                        aliases[0][: -len(in_suffix)],
                        NodeConnectionDirection.OUT,
                    )

                assert (
                    pipe_w_orientation[0] in pipes_set
                    or pipe_w_orientation[0] in cables_set
                    or pipe_w_orientation[0] in gas_pipes_set
                )

                connected_pipes[i] = pipe_w_orientation

        # Note that a pipe series can include both hot and cold pipes for
        # QTH models. It is only about figuring out which pipes are
        # related direction-wise.
        # For Heat models, only hot pipes are allowed to be part of pipe
        # series, as the cold part is zero milp by construction.
        if heat_network_model_type == "QTH":
            alias_relation = self.alias_relation
        elif heat_network_model_type == "Heat":
            # There is no proper AliasRelation yet (because there is milp loss in pipes).
            # So we build one, as that is the easiest way to figure out which pipes are
            # connected to each other in series. We do this by making a temporary/shadow
            # discharge (".Q") variable per pipe, as that way we can share the processing
            # logic for determining pipe series with that of QTH models.
            alias_relation = AliasRelation()

            # Look for aliases only in the hot pipes. All cold pipes are zero by convention anyway.
            hot_pipes = self.hot_pipes.copy()

            pipes_map = {f"{pipe}.HeatIn.Heat": pipe for pipe in hot_pipes}
            pipes_map.update({f"{pipe}.HeatOut.Heat": pipe for pipe in hot_pipes})

            for p in hot_pipes:
                for port in ["In", "Out"]:
                    heat_port = f"{p}.Heat{port}.Heat"
                    connected = self.alias_relation.aliases(heat_port).intersection(
                        pipes_map.keys()
                    )
                    connected.remove(heat_port)

                    if connected:
                        other_pipe_port = next(iter(connected))
                        if other_pipe_port.endswith(f".Heat{port}.Heat"):
                            sign_prefix = "-"
                        else:
                            sign_prefix = ""
                        other_pipe = pipes_map[other_pipe_port]
                        if f"{other_pipe}.Q" not in alias_relation.canonical_variables:
                            alias_relation.add(f"{p}.Q", f"{sign_prefix}{other_pipe}.Q")

        canonical_pipe_qs = {p: alias_relation.canonical_signed(f"{p}.Q") for p in pipes}
        # Move sign from canonical to alias
        canonical_pipe_qs = {(p, d): c for p, (c, d) in canonical_pipe_qs.items()}
        # Reverse the dictionary from `Dict[alias, canonical]` to `Dict[canonical, Set[alias]]`
        pipe_sets = {}
        for a, c in canonical_pipe_qs.items():
            pipe_sets.setdefault(c, []).append(a)

        pipe_series_with_orientation = list(pipe_sets.values())

        # Check that all pipes in the series have the same orientation
        pipe_series = []
        for ps in pipe_series_with_orientation:
            if not len({orientation for _, orientation in ps}) == 1:
                raise Exception(f"Pipes in series {ps} do not all have the same orientation")
            pipe_series.append([name for name, _ in ps])

        buffer_connections = {}

        for b in buffers:
            buffer_connections[b] = []

            for k in ["In", "Out"]:
                b_conn = f"{b}.{heat_network_model_type}{k}"
                prop = "T" if heat_network_model_type == "QTH" else "Heat"
                aliases = [
                    x
                    for x in self.alias_relation.aliases(f"{b_conn}.{prop}")
                    if not x.startswith(b) and x.endswith(f".{prop}")
                ]

                if len(aliases) > 1:
                    raise Exception(f"More than one connection to {b_conn}")
                elif len(aliases) == 0:
                    raise Exception(f"Found no connection to {b_conn}")

                in_suffix = ".QTHIn.T" if heat_network_model_type == "QTH" else ".HeatIn.Heat"
                out_suffix = ".QTHOut.T" if heat_network_model_type == "QTH" else ".HeatOut.Heat"
                alias = aliases[0]
                if alias.endswith(out_suffix):
                    pipe_w_orientation = (
                        alias[: -len(out_suffix)],
                        NodeConnectionDirection.IN,
                    )
                else:
                    assert alias.endswith(in_suffix)
                    pipe_w_orientation = (
                        alias[: -len(in_suffix)],
                        NodeConnectionDirection.OUT,
                    )

                assert pipe_w_orientation[0] in pipes_set

                if k == "In":
                    assert self.is_hot_pipe(pipe_w_orientation[0])
                else:
                    assert self.is_cold_pipe(pipe_w_orientation[0])

                buffer_connections[b].append(pipe_w_orientation)

            buffer_connections[b] = tuple(buffer_connections[b])

        ates_connections = {}

        for a in atess:
            ates_connections[a] = []

            for k in ["In", "Out"]:
                a_conn = f"{a}.{heat_network_model_type}{k}"
                prop = "T" if heat_network_model_type == "QTH" else "Heat"
                aliases = [
                    x
                    for x in self.alias_relation.aliases(f"{a_conn}.{prop}")
                    if not x.startswith(a) and x.endswith(f".{prop}")
                ]

                if len(aliases) > 1:
                    raise Exception(f"More than one connection to {a_conn}")
                elif len(aliases) == 0:
                    raise Exception(f"Found no connection to {a_conn}")

                in_suffix = ".QTHIn.T" if heat_network_model_type == "QTH" else ".HeatIn.Heat"
                out_suffix = ".QTHOut.T" if heat_network_model_type == "QTH" else ".HeatOut.Heat"

                if aliases[0].endswith(out_suffix):
                    pipe_w_orientation = (
                        aliases[0][: -len(out_suffix)],
                        NodeConnectionDirection.IN,
                    )
                else:
                    assert aliases[0].endswith(in_suffix)
                    pipe_w_orientation = (
                        aliases[0][: -len(in_suffix)],
                        NodeConnectionDirection.OUT,
                    )

                assert pipe_w_orientation[0] in pipes_set

                if k == "Out":
                    assert self.is_cold_pipe(pipe_w_orientation[0])
                else:
                    assert self.is_hot_pipe(pipe_w_orientation[0])

                ates_connections[a].append(pipe_w_orientation)

            ates_connections[a] = tuple(ates_connections[a])

        self.__topology = Topology(
            node_connections,
            gas_node_connections,
            pipe_series,
            buffer_connections,
            ates_connections,
            bus_connections,
        )

        super().pre()

    @property
    def energy_system_components(self) -> Dict[str, Set[str]]:
        """
        This method return a dict with the milp network assets ordered per asset type.
        """
        if self.__hn_component_types is None:
            # Create the dictionary once after that it will be available
            string_parameters = self.string_parameters(0)

            # Find the components in model, detection by string
            # (name.component_type: type)
            component_types = sorted({v for k, v in string_parameters.items()})

            components = {}
            for c in component_types:
                components[c] = sorted(
                    {k.split(".")[0] for k, v in string_parameters.items() if v == c}
                )

            self.__hn_component_types = components

        return self.__hn_component_types

    @property
    def energy_system_topology(self) -> Topology:
        """
        This method returns the topology object of the milp network. Which contains specific assets
        with directions on the ports that are needed in the constraints.
        """
        return self.__topology



import logging
import sys
from datetime import timedelta as td

import esdl
from esdl.profiles.influxdbprofilemanager import InfluxDBProfileManager
from esdl.units.conversion import ENERGY_IN_J, POWER_IN_W, convert_to_unit


import pandas as pd

logger = logging.getLogger()

data_set = {}
influx_cred_map = {
    "wu-profiles.esdl-beta.hesi.energy:443": ("warmingup", "warmingup"),
    "omotes-poc-test.hesi.energy:8086": ("write-user", "nwn_write_test"),
}
time_step = td(seconds=3600)
time_step_notation = "{}s".format(int(time_step.total_seconds()))


def parse_esdl_profiles(es, start_date=None, end_date=None):
    logger.info("Caching profiles...")
    error_neighbourhoods = list()
    for profile in [x for x in es.eAllContents() if isinstance(x, esdl.InfluxDBProfile)]:
        profile_host = profile.host
        try:
            # profile associated to asset
            containing_asset_id = profile.eContainer().energyasset.id
        except AttributeError:
            # profile associated to carrier
            containing_asset_id = profile.eContainer().id

        ssl_setting = False
        if "https" in profile_host:
            profile_host = profile_host[8:]
            ssl_setting = True
        elif "http" in profile_host:
            profile_host = profile_host[7:]
        if profile.port == 443:
            ssl_setting = True
        influx_host = "{}:{}".format(profile_host, profile.port)
        if influx_host in influx_cred_map:
            (username, password) = influx_cred_map[influx_host]
        else:
            username = None
            password = None

        time_series_data = InfluxDBProfileManager.create_esdl_influxdb_profile_manager(
            profile,
            username,
            password,
            ssl_setting,
            ssl_setting,
        )

        # Error check start and end dates of profiles
        if time_series_data.end_datetime != profile.endDate:
            logger.error(
                f"The user input profile end datetime: {profile.endDate} does not match the end"
                f" datetime in the datbase: {time_series_data.end_datetime} for variable: "
                f"{profile.field}"
            )
            sys.exit(1)
        if time_series_data.start_datetime != profile.startDate:
            logger.error(
                f"The user input profile start datetime: {profile.startDate} does not match the"
                f" start date in the datbase: {time_series_data.start_datetime} for variable: "
                f"{profile.field}"
            )
            sys.exit(1)
        if time_series_data.start_datetime != time_series_data.profile_data_list[0][0]:
            logger.error(
                f"The profile's variable value for the start datetime: "
                f"{time_series_data.start_datetime} does not match the start datetime of the"
                f" profile data: {time_series_data.profile_data_list[0][0]}"
            )
            sys.exit(1)
        if time_series_data.end_datetime != time_series_data.profile_data_list[-1][0]:
            logger.error(
                f"The profile's variable value for the end datetime: "
                f"{time_series_data.end_datetime} does not match the end datetime of the"
                f" profile data: {time_series_data.profile_data_list[-1][0]}"
            )
            sys.exit(1)

        # Error check: ensure that the profile data has a time resolutuon of 3600s (1hour) as
        # expected
        for idp in range(len(time_series_data.profile_data_list) - 1):
            time_resolution = (
                time_series_data.profile_data_list[idp + 1][0]
                - time_series_data.profile_data_list[idp][0]
            )
            if time_resolution.seconds != 3600:
                logger.error(
                    f"The time resolution of the profile:{profile.measurement}-{profile.field} is"
                    "not 3600s as expected"
                )
                sys.exit(1)

        data_points = {
            t[0].strftime("%Y-%m-%dT%H:%M:%SZ"): t[1] for t in time_series_data.profile_data_list
        }
        df = pd.DataFrame.from_dict(data_points, orient="index")
        df.index = pd.to_datetime(df.index, utc=True)

        # TODO add test case. Currently no test case for esdl parsing
        # Convert Power and Energy to standard unit of Watt and Joules
        for idf in range(len(df)):
            try:
                unit = profile.profileQuantityAndUnit.reference.physicalQuantity
            except AttributeError:
                unit = profile.profileQuantityAndUnit.physicalQuantity
            if unit == esdl.PhysicalQuantityEnum.POWER:
                df.iloc[idf] = convert_to_unit(
                    df.iloc[idf], profile.profileQuantityAndUnit, POWER_IN_W
                )
            elif unit == esdl.PhysicalQuantityEnum.ENERGY:
                df.iloc[idf] = convert_to_unit(
                    df.iloc[idf], profile.profileQuantityAndUnit, ENERGY_IN_J
                )
            elif unit == esdl.PhysicalQuantityEnum.COST:
                # we assume no unit change for now
                pass
            else:
                print(
                    f"Current the code only caters for: {esdl.PhysicalQuantityEnum.POWER} & "
                    f"{esdl.PhysicalQuantityEnum.ENERGY}, and it does not cater for "
                    f"{unit}"
                )
                sys.exit(1)

        data_set[containing_asset_id] = df * profile.multiplier

        if len(error_neighbourhoods) > 0:
            raise RuntimeError(f"Encountered errors loading data for {error_neighbourhoods}")

    return data_set



from mesido.pycml import Variable

from numpy import nan, pi

from ._non_storage_component import _NonStorageComponent


class HeatPipe(_NonStorageComponent):
    """
    The pipe component is to model the pressure drop (and optionally hydraulic power) and
    heat losses over a pipe. Three options for head loss computation are available in the HeatMixin
    options: No_HeadLoss, Linear, DW_Linearized. The hydraulic power computation can only be done
    reasonably if DW_linearized is selected as otherwise head_losses are significantly
    over-estimated.

    The heat to discharge constraints are set in the HeatMixin. Where we ensure that the heat must
    be smaller than the flow can carry, as we overestimate the heat losses with the outgoing
    temperature at the demand/source/storage assets where in reality this temperature drops
    throughout the network. Meaning that the flow does lose energy but not temperature. In this
    manner the energy losses will always be overestimated as in reality the
    flow will also have a temperature drop.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "heat_pipe"
        self.disconnectable = False
        self.has_control_valve = False

        self.length = 1.0
        self.diameter = 1.0
        assert "area" not in modifiers, "modifying area directly is not allowed"
        self.area = 0.25 * pi * self.diameter**2
        self.temperature = nan
        self.carrier_id = -1
        self.pressure = 16.0e5

        # Parameters determining the heat loss
        # All of these have default values in the library function
        self.insulation_thickness = nan
        self.conductivity_insulation = nan
        self.conductivity_subsoil = nan
        self.depth = nan
        self.h_surface = nan
        self.pipe_pair_distance = nan

        self.T_ground = 10.0

        self.Heat_loss = nan

        self.add_variable(Variable, "dH")

        # rho * ff * length * area / 2 / diameter * velocity**3
        ff = 0.02  # Order of magnitude expected with 0.05-2.5m/s in 20mm-1200mm diameter pipe
        velo = self.Q_nominal / self.area
        self.Hydraulic_power_nominal = (
            self.rho * ff * max(self.length, 1.0) * pi * self.area / self.diameter / 2.0 * velo**3
        )
        self.add_variable(
            Variable, "Hydraulic_power", min=0.0, nominal=self.Hydraulic_power_nominal
        )  # [W]

        self.add_equation(
            (self.Hydraulic_power - (self.HeatIn.Hydraulic_power - self.HeatOut.Hydraulic_power))
            / (self.nominal_pressure * self.Q_nominal * self.Hydraulic_power_nominal) ** 0.5
        )

        self.add_equation(((self.Heat_flow - self.HeatIn.Heat) / self.Heat_nominal))

        # Note: Heat loss is added in the mixin, because it depends on the flow direction
        # * heat loss equation: (HeatOut.Heat - (HeatIn.Heat +/- Heat_loss)) = 0.



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestAbsoluteHeat(TestCase):
    def test_absolute_heat(self):
        """
        This is a single line ring model, meaning that there are no dedicated supply or return
        lines. This means that this model pipes are not related (no relation between hot and cold
        pipes exists).

        Checks:
        1. demand is matched
        2. energy conservation in the network
        3. heat to discharge

        """
        import models.absolute_heat.src.example as example
        from models.absolute_heat.src.example import HeatProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="absolute_heat.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )

        demand_matching_test(heat_problem, heat_problem.extract_results())
        energy_conservation_test(heat_problem, heat_problem.extract_results())
        heat_to_discharge_test(heat_problem, heat_problem.extract_results())


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestAbsoluteHeat()
    a.test_absolute_heat()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestProducerMaxProfile(TestCase):
    """
    A test to verify that the producer can have a given scaled profile, where the producer will
    always produce equal or less than said profile. This constraint is checked for a producer,
    where the producer's profile was also intentionally reduced for a couple of time-steps
    (reducing the profile value at a few time steps).

    Checks:
    - Standard checks demand matching, energy conservation and heat to discharge
    - check that heat_source <= scaled_profile * size_source.

    """

    def test_max_producer_profile(self):
        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import HeatProblemProdProfile

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            HeatProblemProdProfile,
            base_folder=base_folder,
            esdl_file_name="3a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )
        results = solution.extract_results()

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)
        tol = 1e-8
        heat_producer = results["GeothermalSource_b702.Heat_source"]
        size_producer = results["GeothermalSource_b702__max_size"]

        heat_producer_profile_scaled = solution.get_timeseries(
            "GeothermalSource_b702.maximum_heat_source"
        ).values
        heat_producer_profile_full = heat_producer_profile_scaled * size_producer

        # check that heat produced is smaller than the profile
        biggerthen = all(heat_producer_profile_full + tol >= heat_producer)
        self.assertTrue(biggerthen)



import math
from typing import List, Optional, Tuple, Union

import numpy as np


def heat_loss_u_values_pipe(
    inner_diameter: float,
    insulation_thicknesses: Union[float, List[float], np.ndarray] = None,
    conductivities_insulation: Union[float, List[float], np.ndarray] = 0.033,
    conductivity_subsoil: float = 2.3,
    depth: float = 1.0,
    h_surface: float = 15.4,
    pipe_distance: float = None,
    neighbour: bool = True,
) -> Tuple[float, float]:
    """
    Calculate the U_1 and U_2 milp loss values for a pipe based for either
    single- or multi-layer insulation. The milp loss calculation for two parallel pipes in the
    ground is based on the literature of Benny Böhm:
        - Original reference (paid article): B. Bohm, On transient milp losses from buried district
        heating pipes, International Journal of Energy Research 24, 1311 (2000))
        - Used in Master's degree: Jort de Boer, Optimization of a District Heating Network with the
        focus on milp loss, TU Delft Mechanical, Maritime and Materials Engineering, 2018,
        http://resolver.tudelft.nl/uuid:7be9fcdd-49e4-4e0c-b36c-69d8b713a874 (access to pdf)

    If the `insulation_thicknesses` is provided as a list, the length should be
    equal to the length of `conductivities_insulation`. If both are floats, a
    single layer of insulation is assumed.

    :inner_diameter:            Inner diameter of the pipes [m]
    :insulation_thicknesses:    Thicknesses of the insulation [m]
                                Default of None means a thickness of 0.5 * inner diameter.
    :conductivities_insulation: Thermal conductivities of the insulation layers [W/m/K]
    :conductivity_subsoil:      Subsoil thermal conductivity [W/m/K]
    :h_surface:                 Heat transfer coefficient at surface [W/m^2/K]
    :param depth:               Depth of outer top of the pipeline [m]
    :param pipe_distance:       Distance between pipeline feed and return pipeline centers [m].
                                Default of None means 2 * outer diameter

    :return: U-values (U_1 / U_2) for milp losses of pipes [W/(m*K)]
    """

    if insulation_thicknesses is None:
        insulation_thicknesses = 0.5 * inner_diameter

    if not isinstance(insulation_thicknesses, type(conductivities_insulation)):
        raise Exception("Insulation thicknesses and conductivities should have the same type.")

    if hasattr(insulation_thicknesses, "__iter__"):
        if not len(insulation_thicknesses) == len(conductivities_insulation):
            raise Exception(
                "Number of insulation thicknesses should match number of conductivities"
            )
        insulation_thicknesses = np.array(insulation_thicknesses)
        conductivities_insulation = np.array(conductivities_insulation)
    else:
        insulation_thicknesses = np.array([insulation_thicknesses])
        conductivities_insulation = np.array([conductivities_insulation])

    diam_inner = inner_diameter
    diam_outer = diam_inner + 2 * sum(insulation_thicknesses)
    if pipe_distance is None:
        pipe_distance = 2 * diam_outer
    depth_center = depth + 0.5 * diam_outer

    # NOTE: We neglect the milp resistance due to convection inside the pipe,
    # i.e. we assume perfect mixing, or that this resistance is much lower
    # than the resistance of the outer insulation layers.

    # Heat resistance of the subsoil
    r_subsoil = 1 / (2 * math.pi * conductivity_subsoil) * math.log(4.0 * depth_center / diam_outer)

    # Heat resistance due to insulation
    outer_diameters = diam_inner + 2.0 * np.cumsum(insulation_thicknesses)
    inner_diameters = np.array([inner_diameter, *outer_diameters[:-1]])
    r_ins = sum(
        np.log(outer_diameters / inner_diameters) / (2.0 * math.pi * conductivities_insulation)
    )

    # Heat resistance due to neighboring pipeline
    r_m = (
        1
        / (4 * math.pi * conductivity_subsoil)
        * math.log(1 + (2 * depth_center / pipe_distance) ** 2)
    )

    if neighbour:
        u_1 = (r_subsoil + r_ins) / ((r_subsoil + r_ins) ** 2 - r_m**2)
        u_2 = r_m / ((r_subsoil + r_ins) ** 2 - r_m**2)
    else:
        u_1 = 1 / (r_subsoil + r_ins)
        u_2 = 0

    return u_1, u_2


def pipe_heat_loss(
    optimization_problem,
    options,
    parameters,
    p: str,
    u_values: Optional[Tuple[float, float]] = None,
    temp: float = None,
):
    """
    The milp losses have three components:

    - dependency on the pipe temperature
    - dependency on the ground temperature
    - dependency on temperature difference between the supply/return line.

    This latter term assumes that the supply and return lines lie close
    to, and thus influence, each other. I.e., the supply line loses milp
    that is absorbed by the return line. Note that the term dtemp is
    positive when the pipe is in the supply line and negative otherwise.
    """
    if options["neglect_pipe_heat_losses"]:
        return 0.0

    neighbour = optimization_problem.has_related_pipe(p)

    if u_values is None:
        u_kwargs = {
            "inner_diameter": parameters[f"{p}.diameter"],
            "insulation_thicknesses": parameters[f"{p}.insulation_thickness"],
            "conductivities_insulation": parameters[f"{p}.conductivity_insulation"],
            "conductivity_subsoil": parameters[f"{p}.conductivity_subsoil"],
            "depth": parameters[f"{p}.depth"],
            "h_surface": parameters[f"{p}.h_surface"],
            "pipe_distance": parameters[f"{p}.pipe_pair_distance"],
        }

        # NaN values mean we use the function default
        u_kwargs = {k: v for k, v in u_kwargs.items() if not np.all(np.isnan(v))}
        u_1, u_2 = heat_loss_u_values_pipe(**u_kwargs, neighbour=neighbour)
    else:
        u_1, u_2 = u_values

    length = parameters[f"{p}.length"]
    temperature = parameters[f"{p}.temperature"]
    if temp is not None:
        temperature = temp
    temperature_ground = parameters[f"{p}.T_ground"]
    if neighbour:
        if optimization_problem.is_hot_pipe(p):
            dtemp = (
                temperature - parameters[f"{optimization_problem.hot_to_cold_pipe(p)}.temperature"]
            )
        else:
            dtemp = (
                temperature - parameters[f"{optimization_problem.cold_to_hot_pipe(p)}.temperature"]
            )
    else:
        dtemp = 0

    # if no return/supply pipes can be linked to eachother, the influence of the milp of the
    # neighbouring pipes can also not be determined and thus no influence is assumed
    # (distance between pipes to infinity)
    # This results in Rneighbour -> 0 and therefore u2->0, u1-> 1/(Rsoil+Rins)

    heat_loss = (
        length * (u_1 - u_2) * temperature
        - (length * (u_1 - u_2) * temperature_ground)
        + (length * u_2 * dtemp)
    )

    if heat_loss < 0 and temperature > temperature_ground:
        raise Exception(f"Heat loss of pipe {p} should be nonnegative.")

    return heat_loss



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestHEX(TestCase):
    def test_heat_exchanger(self):
        """
        Check the modelling of the heat exchanger component which allows two hydraulically
        decoupled networks to exchange heat with each other. It is enforced that heat can only flow
        from the primary side to the secondary side, and heat exchangers are allowed to be disabled
        for timesteps in which they are not used. This is to allow for the temperature constraints
        (T_primary > T_secondary) to become deactivated.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation
        - That the efficiency is correclty implemented for heat from primary to secondary
        - Check that the is_disabled is set correctly.
        - Check if the temperatures provided are physically feasible.

        """
        import models.heat_exchange.src.run_heat_exchanger as run_heat_exchanger
        from models.heat_exchange.src.run_heat_exchanger import (
            HeatProblem,
        )

        base_folder = Path(run_heat_exchanger.__file__).resolve().parent.parent
        # -----------------------------------------------------------------------------------------
        # Do not delete: this is used to manualy check writing out of profile data

        class HeatProblemPost(HeatProblem):
            # def post(self):
            #     super().post()
            #     self._write_updated_esdl(
            #         self._ESDLMixin__energy_system_handler.energy_system,
            #         optimizer_sim=True,
            #     )

            def energy_system_options(self):
                options = super().energy_system_options()
                # self.heat_network_settings["minimize_head_losses"] = True  # used for manual tests
                return options

        # Do not delete kwargs: this is used to manualy check writing out of profile data
        kwargs = {
            "write_result_db_profiles": False,
            "influxdb_host": "localhost",
            "influxdb_port": 8086,
            "influxdb_username": None,
            "influxdb_password": None,
            "influxdb_ssl": False,
            "influxdb_verify_ssl": False,
        }
        # -----------------------------------------------------------------------------------------

        solution = run_esdl_mesido_optimization(
            HeatProblemPost,
            base_folder=base_folder,
            esdl_file_name="heat_exchanger.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
            **kwargs,
        )

        results = solution.extract_results()
        parameters = solution.parameters(0)

        prim_heat = results["HeatExchange_39ed.Primary_heat"]
        sec_heat = results["HeatExchange_39ed.Secondary_heat"]
        disabled = results["HeatExchange_39ed__disabled"]

        # We check the energy converted betweeen the commodities
        eff = parameters["HeatExchange_39ed.efficiency"]

        demand_matching_test(solution, results)
        heat_to_discharge_test(solution, results)
        energy_conservation_test(solution, results)

        np.testing.assert_allclose(prim_heat * eff, sec_heat)

        # Note that we are not testing the last element as we exploit the last timestep for
        # checking the disabled boolean and the assert statement doesn't work for a difference of
        # zero
        np.testing.assert_allclose(prim_heat[-1], 0.0, atol=1e-5)
        np.testing.assert_allclose(disabled[-1], 1.0)
        np.testing.assert_allclose(disabled[:-1], 0.0)
        # Check that heat is flowing through the hex
        np.testing.assert_array_less(-prim_heat[:-1], 0.0)

        np.testing.assert_array_less(
            parameters["HeatExchange_39ed.Secondary.T_supply"],
            parameters["HeatExchange_39ed.Primary.T_supply"],
        )
        np.testing.assert_array_less(
            parameters["HeatExchange_39ed.Secondary.T_return"],
            parameters["HeatExchange_39ed.Primary.T_return"],
        )


class TestHP(TestCase):
    def test_heat_pump(self):
        """
        Check the modelling of the heat pump component which has a constant COP with no energy loss.
        In this specific problem we expect the use of the secondary source to be maximised as
        electrical heat from the HP is "free".

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation
        - Check that the heat pump is producing according to its COP
        - Check that Secondary source use in minimized


        """
        import models.heatpump.src.run_heat_pump as run_heat_pump
        from models.heatpump.src.run_heat_pump import (
            HeatProblem,
        )

        base_folder = Path(run_heat_pump.__file__).resolve().parent.parent

        # -----------------------------------------------------------------------------------------
        # Do not delete: this is used to manualy check writing out of profile data

        class HeatProblemPost(HeatProblem):
            # def post(self):
            #     super().post()
            #     self._write_updated_esdl(
            #         self._ESDLMixin__energy_system_handler.energy_system,
            #         optimizer_sim=True,
            #     )

            def energy_system_options(self):
                options = super().energy_system_options()
                # self.heat_network_settings["minimize_head_losses"] = True  # used for manual tests
                return options

        # Do not delete kwargs: this is used to manualy check writing out of profile data
        kwargs = {
            "write_result_db_profiles": False,
            "influxdb_host": "localhost",
            "influxdb_port": 8086,
            "influxdb_username": None,
            "influxdb_password": None,
            "influxdb_ssl": False,
            "influxdb_verify_ssl": False,
        }
        # -----------------------------------------------------------------------------------------

        solution = run_esdl_mesido_optimization(
            HeatProblemPost,
            base_folder=base_folder,
            esdl_file_name="heat_pump.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
            **kwargs,
        )

        results = solution.extract_results()
        parameters = solution.parameters(0)

        prim_heat = results["GenericConversion_3d3f.Primary_heat"]
        sec_heat = results["GenericConversion_3d3f.Secondary_heat"]
        power_elec = results["GenericConversion_3d3f.Power_elec"]

        # Check that only the minimum velocity is flowing through the secondary source.
        cross_sectional_area = parameters["Pipe3.area"]
        np.testing.assert_allclose(
            results["ResidualHeatSource_aec9.Q"] / cross_sectional_area,
            1.0e-3,
            atol=2.5e-5,
        )

        demand_matching_test(solution, results)
        heat_to_discharge_test(solution, results)
        energy_conservation_test(solution, results)

        # We check the energy converted betweeen the commodities
        np.testing.assert_allclose(power_elec * parameters["GenericConversion_3d3f.COP"], sec_heat)
        np.testing.assert_allclose(power_elec + prim_heat, sec_heat)


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestHEX()
    a.test_heat_exchanger()

    b = TestHP()
    b.test_heat_pump()



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestMaxSizeAggregationCount(TestCase):
    def test_max_size_and_aggr_count(self):
        """
        Check the behaviour of the asset sizing, asset placement and the cost
        components associated to those variables.

        - the asset size should be at least as large as it's maximum utilization
        - the asset should be placed when it is utilized
        - the cost components should be present if the asset is placed and used.
        - For this scenario a problem is set-up with two optional sources (with their connection
        pipes) where one source can supply the network by itself. It is expected that only one
        source will be placed due to the minimization of the cost (which cost is this?, I assume
        operational cost) and installation cost. The placement behaviour is further tested in a
        second case by adding an optional ates and buffer. However, these 2 additional optional
        assets should not be placed by the optmizer because of heat losses.

        Checks:
        - Check that source 1 is utilized and also placed
        - Check that source 2 is utilized and placed
        - Check that max_size source 1 is zero
        - Check that max_size source 2 is > utilization
        - Check cost components for source 1 and 2
        - Check max size and placement of ates
        - Fixed operational cost for sources

        """
        import models.test_case_small_network_with_ates_with_buffer.src.run_ates as run_ates
        from models.test_case_small_network_with_ates_with_buffer.src.run_ates import (
            HeatProblem,
        )

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over a few days
        solution = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        results = solution.extract_results()
        parameters = solution.parameters(0)

        # Producer 1 should not produce due to higher cost
        # Producer 2 should produce
        heat_1 = results["HeatProducer_1.Heat_source"]
        heat_2 = results["HeatProducer_2.Heat_source"]
        prod_1_placed = results["HeatProducer_1_aggregation_count"]
        prod_2_placed = results["HeatProducer_2_aggregation_count"]
        var_cost_1 = results["HeatProducer_1__variable_operational_cost"]
        var_cost_2 = results["HeatProducer_2__variable_operational_cost"]
        fix_cost_1 = results["HeatProducer_1__fixed_operational_cost"]
        fix_cost_2 = results["HeatProducer_2__fixed_operational_cost"]
        inst_cost_1 = results["HeatProducer_1__installation_cost"]
        inst_cost_2 = results["HeatProducer_2__installation_cost"]
        inv_cost_1 = results["HeatProducer_1__investment_cost"]
        inv_cost_2 = results["HeatProducer_2__investment_cost"]
        max_size_1 = results["HeatProducer_1__max_size"]
        max_size_2 = results["HeatProducer_2__max_size"]

        # Test if source 1 is off and 2 is producing
        np.testing.assert_allclose(heat_1, 0.0, atol=1e-6)
        np.testing.assert_equal(True, heat_2[1:] > 0.0)

        # Test if source 1 is not placed and 2 is placed
        np.testing.assert_allclose(prod_1_placed, 0.0)
        np.testing.assert_allclose(prod_2_placed, 1.0, atol=1.0e-6)

        # Test that max size is correct, note that we use an equality check as due to the cost
        # minimization they should be equal.
        np.testing.assert_allclose(max(heat_2), max_size_2)
        np.testing.assert_allclose(max_size_1, 0.0)

        # Test that investmentcost is correctly linked to max size
        np.testing.assert_allclose(
            inv_cost_2,
            solution.parameters(0)["HeatProducer_2.investment_cost_coefficient"] * max_size_2,
        )

        # Test that cost only exist for 2 and not for 1. Note the tolerances
        # to avoid test failing when heat losses slightly change
        np.testing.assert_allclose(var_cost_1, 0.0, atol=1e-9)
        np.testing.assert_allclose(
            var_cost_2,
            np.sum(
                results["HeatProducer_2.Heat_source"][1:]
                * (solution.times()[1:] - solution.times()[:-1])
                / 3600
            )
            * parameters["HeatProducer_2.variable_operational_cost_coefficient"],
            atol=1000.0,
            rtol=1.0e-2,
        )
        np.testing.assert_allclose(fix_cost_1, 0.0, atol=1.0e-6)
        np.testing.assert_allclose(
            fix_cost_2,
            max_size_2 * parameters["HeatProducer_2.fixed_operational_cost_coefficient"],
            atol=1.0e-6,
        )
        np.testing.assert_allclose(inst_cost_1, 0.0, atol=1e-9)
        np.testing.assert_allclose(inst_cost_2, 100000.0)
        np.testing.assert_allclose(inv_cost_1, 0.0, atol=1e-9)
        np.testing.assert_allclose(
            inv_cost_2,
            max_size_2 * parameters["HeatProducer_2.investment_cost_coefficient"],
            atol=1.0,
            rtol=1.0e-2,
        )

        # Since the buffer and ates are not optional they must consume some heat to compensate
        # losses as the buffer has a minimum fraction volume of 5%.
        # Therefore, we can check the max_size constraint.
        np.testing.assert_allclose(
            True, results["HeatStorage_74c1.Stored_heat"] <= results["HeatStorage_74c1__max_size"]
        )
        np.testing.assert_allclose(
            True, abs(results["ATES_033c.Heat_ates"]) <= results["ATES_033c__max_size"]
        )
        np.testing.assert_allclose(results["ATES_033c_aggregation_count"], 1.0)
        np.testing.assert_allclose(results["HeatStorage_74c1_aggregation_count"], 1.0)

        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates
        from models.test_case_small_network_ates_buffer_optional_assets.src.run_ates import (
            HeatProblem,
        )

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is the same problem, but now with the buffer and ates also optional.
        # Therefore, we expect that the ates and buffer are no longer placed to avoid their heat
        # losses. This allows us to check if their placement constraints are proper.
        solution = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates_with_buffer_all_optional.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        results = solution.extract_results()

        np.testing.assert_allclose(results["ATES_033c.Heat_ates"], 0.0, atol=1.0e-6)
        np.testing.assert_allclose(results["HeatStorage_74c1.Stored_heat"], 0.0, atol=1.0e-3)
        np.testing.assert_allclose(results["ATES_033c_aggregation_count"], 0.0, atol=1.0e-6)
        np.testing.assert_allclose(results["HeatStorage_74c1_aggregation_count"], 0.0, atol=1.0e-6)

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)



from .buffer import Buffer
from .check_valve import CheckValve
from .control_valve import ControlValve
from .demand import Demand
from .geothermal_source import GeothermalSource
from .node import Node
from .pipe import Pipe
from .pump import Pump
from .qth_port import QTHPort
from .qth_two_port import QTHTwoPort
from .source import Source

__all__ = [
    "Buffer",
    "CheckValve",
    "ControlValve",
    "Demand",
    "GeothermalSource",
    "Node",
    "Pipe",
    "Pump",
    "QTHPort",
    "QTHTwoPort",
    "Source",
]



from mesido.pycml import Variable

from numpy import nan

from ._non_storage_component import _NonStorageComponent


class Pump(_NonStorageComponent):
    """
    The pump component is there to add head to the flow. We assume head can only be added for
    positive flow.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.temperature = nan
        self.carrier_id = -1

        self.component_type = "pump"

        self.pump_efficiency = 0.5

        self.add_variable(Variable, "dH", min=0.0)
        self.add_variable(
            Variable, "Pump_power", min=0.0, nominal=self.Q_nominal * self.nominal_pressure
        )

        self.add_equation(self.dH - (self.HeatOut.H - self.HeatIn.H))

        self.add_equation(
            (self.Pump_power - (self.HeatOut.Hydraulic_power - self.HeatIn.Hydraulic_power))
            / (self.Q_nominal * self.nominal_pressure)
        )

        self.add_equation((self.HeatIn.Heat - self.HeatOut.Heat) / self.Heat_nominal)



from mesido.pycml import Variable
from mesido.pycml.component_library.milp._internal import BaseAsset
from mesido.pycml.component_library.milp._internal.electricity_component import (
    ElectricityComponent,
)
from mesido.pycml.component_library.milp.electricity.electricity_base import (
    ElectricityPort,
)
from mesido.pycml.component_library.milp.gas.gas_base import (
    GasPort,
)

from numpy import nan


class Electrolyzer(ElectricityComponent, BaseAsset):
    """
    An electrolyzer consumes electricity and produces hydrogen
    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(),
                modifiers,
            ),
        )

        self.component_type = "electrolyzer"

        self.a_eff_coefficient = nan
        self.b_eff_coefficient = nan
        self.c_eff_coefficient = nan

        self.minimum_load = nan

        self.density = 2.5  # H2 density [kg/m3] at 30bar

        self.Q_nominal = nan
        self.min_voltage = nan

        self.efficiency = nan

        self.nominal_gass_mass_out = self.Q_nominal * self.density
        self.nominal_power_consumed = nan

        self.add_variable(ElectricityPort, "ElectricityIn")  # [W]
        self.add_variable(Variable, "Power_consumed", min=0.0, nominal=self.nominal_power_consumed)
        self.add_equation(
            (self.ElectricityIn.Power - self.Power_consumed) / self.nominal_power_consumed
        )  # [W]

        self.add_variable(GasPort, "GasOut")
        self.add_variable(
            Variable, "Gas_mass_flow_out", min=0.0, nominal=self.nominal_gass_mass_out
        )  # [g/s]
        self.add_equation(
            (self.GasOut.mass_flow - self.Gas_mass_flow_out) / self.nominal_gass_mass_out
        )
